import asyncio
from asyncio import Task
from dataclasses import dataclass, field
from functools import lru_cache
from typing import Any, Optional

import jq  # type: ignore
from loguru import logger

from port_ocean.context.ocean import ocean
from port_ocean.core.handlers.entity_processor.base import BaseEntityProcessor
from port_ocean.core.handlers.port_app_config.models import ResourceConfig
from port_ocean.core.models import Entity
from port_ocean.core.ocean_types import (
    RAW_ITEM,
    EntitySelectorDiff,
    CalculationResult,
)
from port_ocean.core.utils import gather_and_split_errors_from_results, zip_and_sum
from port_ocean.exceptions.core import EntityProcessorException
from port_ocean.utils.queue_utils import process_in_queue


@dataclass
class MappedEntity:
    """Represents the entity after applying the mapping

    This class holds the mapping entity along with the selector boolean value and optionally the raw data.
    """

    entity: dict[str, Any] = field(default_factory=dict)
    did_entity_pass_selector: bool = False
    raw_data: Optional[dict[str, Any]] = None


class JQEntityProcessor(BaseEntityProcessor):
    """Processes and parses entities using JQ expressions.

    This class extends the BaseEntityProcessor and provides methods for processing and
    parsing entities based on PyJQ queries. It supports compiling and executing PyJQ patterns,
    searching for data in dictionaries, and transforming data based on object mappings.
    """

    @lru_cache
    def _compile(self, pattern: str) -> Any:
        if not ocean.config.allow_environment_variables_jq_access:
            pattern = "def env: {}; {} as $ENV | " + pattern
        return jq.compile(pattern)

    @staticmethod
    def _stop_iterator_handler(func: Any) -> Any:
        """
        Wrap the function to handle StopIteration exceptions.
        Prevents StopIteration from stopping the thread and skipping further queue processing.
        """

        def inner() -> Any:
            try:
                return func()
            except StopIteration:
                return None

        return inner

    async def _search(self, data: dict[str, Any], pattern: str) -> Any:
        try:
            loop = asyncio.get_event_loop()
            compiled_pattern = self._compile(pattern)
            func = compiled_pattern.input_value(data)
            return await loop.run_in_executor(
                None, self._stop_iterator_handler(func.first)
            )
        except Exception as exc:
            logger.debug(
                f"Search failed for pattern '{pattern}' in data: {data}, Error: {exc}"
            )
            return None

    async def _search_as_bool(self, data: dict[str, Any], pattern: str) -> bool:
        loop = asyncio.get_event_loop()

        compiled_pattern = self._compile(pattern)
        func = compiled_pattern.input_value(data)

        value = await loop.run_in_executor(
            None, self._stop_iterator_handler(func.first)
        )
        if isinstance(value, bool):
            return value
        raise EntityProcessorException(
            f"Expected boolean value, got value:{value} of type: {type(value)} instead"
        )

    async def _search_as_object(
        self, data: dict[str, Any], obj: dict[str, Any]
    ) -> dict[str, Any | None]:
        search_tasks: dict[
            str, Task[dict[str, Any | None]] | list[Task[dict[str, Any | None]]]
        ] = {}
        for key, value in obj.items():
            if isinstance(value, list):
                search_tasks[key] = [
                    asyncio.create_task(self._search_as_object(data, obj))
                    for obj in value
                ]

            elif isinstance(value, dict):
                search_tasks[key] = asyncio.create_task(
                    self._search_as_object(data, value)
                )
            else:
                search_tasks[key] = asyncio.create_task(self._search(data, value))

        result: dict[str, Any | None] = {}
        for key, task in search_tasks.items():
            try:
                if isinstance(task, list):
                    result[key] = [await task for task in task]
                else:
                    result[key] = await task
            except Exception:
                result[key] = None

        return result

    async def _get_mapped_entity(
        self,
        data: dict[str, Any],
        raw_entity_mappings: dict[str, Any],
        selector_query: str,
        parse_all: bool = False,
    ) -> MappedEntity:
        should_run = await self._search_as_bool(data, selector_query)
        if parse_all or should_run:
            mapped_entity = await self._search_as_object(data, raw_entity_mappings)
            return MappedEntity(
                mapped_entity,
                did_entity_pass_selector=should_run,
                raw_data=data if should_run else None,
            )

        return MappedEntity()

    async def _calculate_entity(
        self,
        data: dict[str, Any],
        raw_entity_mappings: dict[str, Any],
        items_to_parse: str | None,
        selector_query: str,
        parse_all: bool = False,
    ) -> tuple[list[MappedEntity], list[Exception]]:
        raw_data = [data.copy()]
        if items_to_parse:
            items = await self._search(data, items_to_parse)
            if not isinstance(items, list):
                logger.warning(
                    f"Failed to parse items for JQ expression {items_to_parse}, Expected list but got {type(items)}."
                    f" Skipping..."
                )
                return [], []
            raw_data = [{"item": item, **data} for item in items]

        entities, errors = await gather_and_split_errors_from_results(
            [
                self._get_mapped_entity(
                    raw,
                    raw_entity_mappings,
                    selector_query,
                    parse_all,
                )
                for raw in raw_data
            ]
        )
        if errors:
            logger.error(
                f"Failed to calculate entities with {len(errors)} errors. errors: {errors}"
            )
        return entities, errors

    @staticmethod
    async def _send_examples(data: list[dict[str, Any]], kind: str) -> None:
        try:
            if data:
                await ocean.port_client.ingest_integration_kind_examples(
                    kind, data, should_log=False
                )
        except Exception as ex:
            logger.warning(
                f"Failed to send raw data example {ex}",
                exc_info=True,
            )

    async def _parse_items(
        self,
        mapping: ResourceConfig,
        raw_results: list[RAW_ITEM],
        parse_all: bool = False,
        send_raw_data_examples_amount: int = 0,
    ) -> CalculationResult:
        raw_entity_mappings: dict[str, Any] = mapping.port.entity.mappings.dict(
            exclude_unset=True
        )
        logger.info(f"Parsing {len(raw_results)} raw results into entities")
        calculated_entities_results, errors = zip_and_sum(
            await process_in_queue(
                raw_results,
                self._calculate_entity,
                raw_entity_mappings,
                mapping.port.items_to_parse,
                mapping.selector.query,
                parse_all,
            )
        )
        logger.debug(
            f"Finished parsing raw results into entities with {len(errors)} errors. errors: {errors}"
        )

        passed_entities = []
        failed_entities = []
        examples_to_send: list[dict[str, Any]] = []
        for result in calculated_entities_results:
            if result.entity.get("identifier") and result.entity.get("blueprint"):
                parsed_entity = Entity.parse_obj(result.entity)
                if result.did_entity_pass_selector:
                    passed_entities.append(parsed_entity)
                    if (
                        len(examples_to_send) < send_raw_data_examples_amount
                        and result.raw_data is not None
                    ):
                        examples_to_send.append(result.raw_data)
                else:
                    failed_entities.append(parsed_entity)
        if (
            not calculated_entities_results
            and raw_results
            and send_raw_data_examples_amount > 0
        ):
            logger.warning(
                f"No entities were parsed from {len(raw_results)} raw results, sending raw data examples"
            )
            examples_to_send = raw_results[:send_raw_data_examples_amount]

        await self._send_examples(examples_to_send, mapping.kind)

        return CalculationResult(
            EntitySelectorDiff(passed=passed_entities, failed=failed_entities),
            errors,
        )



# -*- coding: utf-8 -*-

import click

from port_ocean import __version__, __integration_version__
from port_ocean.cli.commands.main import cli_start, print_logo, console
from port_ocean.config.settings import LogLevelType


@cli_start.command()
@click.argument("path", default=".", type=click.Path(exists=True))
@click.option(
    "-l",
    "--log-level",
    "log_level",
    type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]),
    default="INFO",
    help="""Set the logging level for the integration.
            Supported levels are DEBUG, INFO, WARNING, ERROR,
            and CRITICAL. If not specified, the default level
            is INFO.""",
)
@click.option(
    "-p",
    "--port",
    "port",
    type=int,
    default=8000,
    help="""Set the port for the integration to run on.
            If not specified, the default port is 8000.""",
)
@click.option(
    "-i",
    "--initialize-port-resources",
    "initialize_port_resources",
    type=bool,
    help="""Set to False to not create default resources on installation.
            If not specified, will use the environment variable `OCEAN__INITIALIZE_PORT_RESOURCES` to determine whether 
            to initialize resources, and if not set, will default to True.""",
)
@click.option(
    "-O",
    "--once",
    "once",
    type=bool,
    is_flag=True,
    help="""specify the option to run the integration once and exit.""",
)
def sail(
    path: str,
    log_level: LogLevelType,
    port: int,
    initialize_port_resources: bool | None,
    once: bool,
) -> None:
    """
    Runs the integration in the given PATH. if no PATH is provided, the current directory will be used.

    PATH: Path to the integration.
    """
    from port_ocean import run

    print_logo()

    console.print("Setting sail... ⛵️⚓️⛵️⚓️ All hands on deck! ⚓️")
    console.print(f"🌊 Ocean version: {__version__}")
    console.print(f"🚢 Integration version: {__integration_version__}")

    override = {}
    if once:
        console.print("Setting event listener to Once")
        override["event_listener"] = {"type": "ONCE"}

    run(
        path,
        log_level,
        port,
        initialize_port_resources=initialize_port_resources,
        config_override=override,
    )



from enum import Enum
from typing import TypedDict


class UserAgentType(Enum):
    exporter = "exporter"
    gitops = "gitops"


KafkaCreds = TypedDict(
    "KafkaCreds",
    {
        "username": str,
        "password": str,
    },
)

RequestOptions = TypedDict(
    "RequestOptions",
    {
        "merge": bool,
        "create_missing_related_entities": bool,
        "delete_dependent_entities": bool,
        "validation_only": bool,
    },
)



from os import environ
from port_ocean.clients.port.client import PortClient

from loguru import logger
from pydantic import BaseModel

from port_ocean.tests.helpers.integration import cleanup_integration
from port_ocean.tests.helpers.port_client import get_port_client_for_integration


class SmokeTestDetails(BaseModel):
    integration_identifier: str
    blueprint_department: str
    blueprint_person: str
    integration_type: str
    integration_version: str


def get_smoke_test_details() -> SmokeTestDetails:
    blueprint_department = "fake-department"
    blueprint_person = "fake-person"
    integration_identifier = "smoke-test-integration"
    smoke_test_suffix = environ.get("SMOKE_TEST_SUFFIX")
    if smoke_test_suffix is not None:
        integration_identifier = f"{integration_identifier}-{smoke_test_suffix}"
        blueprint_person = f"{blueprint_person}-{smoke_test_suffix}"
        blueprint_department = f"{blueprint_department}-{smoke_test_suffix}"

    return SmokeTestDetails(
        integration_identifier=integration_identifier,
        blueprint_person=blueprint_person,
        blueprint_department=blueprint_department,
        integration_version="0.1.4-dev",
        integration_type="smoke-test",
    )


async def cleanup_smoke_test() -> None:
    smoke_test_details = get_smoke_test_details()
    client_id = environ.get("PORT_CLIENT_ID")
    client_secret = environ.get("PORT_CLIENT_SECRET")

    if not client_secret or not client_id:
        assert False, "Missing port credentials"

    base_url = environ.get("PORT_BASE_URL")
    client = get_port_client_for_integration(
        client_id,
        client_secret,
        smoke_test_details.integration_identifier,
        smoke_test_details.integration_type,
        smoke_test_details.integration_version,
        base_url,
    )

    logger.info("Cleaning up fake integration")
    await cleanup_integration(
        client,
        [smoke_test_details.blueprint_department, smoke_test_details.blueprint_person],
    )
    logger.info("Cleaning up fake integration complete")


def get_port_client_for_fake_integration() -> PortClient:
    smoke_test_details = get_smoke_test_details()
    client_id = environ.get("PORT_CLIENT_ID")
    client_secret = environ.get("PORT_CLIENT_SECRET")

    if not client_secret or not client_id:
        assert False, "Missing port credentials"

    base_url = environ.get("PORT_BASE_URL")
    client = get_port_client_for_integration(
        client_id,
        client_secret,
        smoke_test_details.integration_identifier,
        smoke_test_details.integration_type,
        smoke_test_details.integration_version,
        base_url,
    )

    return client



import asyncio
import json
import sys
from asyncio import ensure_future, Task
from typing import Any, Literal

from confluent_kafka import Message  # type: ignore
from loguru import logger

from port_ocean.consumers.kafka_consumer import KafkaConsumer, KafkaConsumerConfig
from port_ocean.context.ocean import (
    ocean,
)
from port_ocean.core.event_listener.base import (
    BaseEventListener,
    EventListenerEvents,
    EventListenerSettings,
)


class KafkaEventListenerSettings(EventListenerSettings):
    """
    This class inherits from `EventListenerSettings`, which provides a foundation for defining event listener configurations.
    The `KafkaEventListenerSettings` specifically includes settings related to the Kafka event listener.

    Attributes:
        type (Literal["KAFKA"]): A literal indicating the type of the event listener, which is set to "KAFKA" for this class.
        brokers (str): The comma-separated list of Kafka broker URLs to connect to.
        security_protocol (str): The security protocol used for communication with Kafka brokers.
                                 The default value is "SASL_SSL".
        authentication_mechanism (str): The authentication mechanism used for secure access to Kafka brokers.
                                        The default value is "SCRAM-SHA-512".
        kafka_security_enabled (bool): A flag indicating whether Kafka security is enabled.
                                       If True, credentials and security settings are used to connect to Kafka.
                                       The default value is True.
        consumer_poll_timeout (int): The maximum time in seconds to wait for messages during a poll.
                                     The default value is 1 second.
    """

    type: Literal["KAFKA"]
    brokers: str = (
        "b-1-public.publicclusterprod.t9rw6w.c1.kafka.eu-west-1.amazonaws.com:9196,b-2-public.publicclusterprod.t9rw6w.c1.kafka.eu-west-1.amazonaws.com:9196,b-3-public.publicclusterprod.t9rw6w.c1.kafka.eu-west-1.amazonaws.com:9196"
    )
    security_protocol: str = "SASL_SSL"
    authentication_mechanism: str = "SCRAM-SHA-512"
    kafka_security_enabled: bool = True
    consumer_poll_timeout: int = 1


class KafkaEventListener(BaseEventListener):
    """
    The `KafkaEventListener` specifically listens for messages from a Kafka consumer related to changes in an integration.

    Parameters:
        events (EventListenerEvents): A dictionary containing event types and their corresponding event handlers.
        event_listener_config (KafkaEventListenerSettings): Configuration settings for the Kafka event listener.
        org_id (str): The identifier of the organization associated with the integration.
        integration_identifier (str): The identifier of the integration being monitored.
        integration_type (str): The type of the integration being monitored.
    """

    def __init__(
        self,
        events: EventListenerEvents,
        event_listener_config: KafkaEventListenerSettings,
        org_id: str,
        integration_identifier: str,
        integration_type: str,
    ):
        super().__init__(events)
        self.event_listener_config = event_listener_config
        self.org_id = org_id
        self.integration_identifier = integration_identifier
        self.integration_type = integration_type
        self._running_task: Task[Any] | None = None
        self.consumer: KafkaConsumer | None = None

    async def _get_kafka_config(self) -> KafkaConsumerConfig:
        """
        A private method that returns the Kafka consumer configuration based on the provided settings.
        If Kafka security is enabled, it fetches Kafka credentials using the ocean.port_client.get_kafka_creds() method.
        Otherwise, it returns the KafkaConsumerConfig object parsed from the event_listener_config.
        """
        if self.event_listener_config.kafka_security_enabled:
            creds = await ocean.port_client.get_kafka_creds()
            return KafkaConsumerConfig(
                **self.event_listener_config.dict(),
                username=creds.get("username"),
                password=creds.get("password"),
                group_name=f"{self.integration_type}.{self.integration_identifier}",
            )

        return KafkaConsumerConfig.parse_obj(self.event_listener_config.dict())

    def _should_be_processed(self, msg_value: dict[Any, Any], topic: str) -> bool:
        after = msg_value.get("diff", {}).get("after", {})
        # handles delete events from change log where there is no after
        if after is None:
            return False

        integration_identifier = after.get("identifier")
        if integration_identifier != self.integration_identifier:
            return False

        if after.get("updatedAt") == after.get("resyncState", {}).get("updatedAt"):
            return False

        if "change.log" in topic:
            return msg_value.get("changelogDestination", {}).get("type", "") == "KAFKA"

        return False

    async def _handle_message(self, raw_msg: Message) -> None:
        """
        A private method that handles incoming Kafka messages.
        If the message should be processed (determined by `_should_be_processed`), it triggers the corresponding event handler.

        Spawning a thread to handle the message allows the Kafka consumer to continue polling for new messages.
        Using wrap_method_with_context ensures that the thread has access to the current context.
        """
        message = json.loads(raw_msg.value().decode())
        topic = raw_msg.topic()

        if not self._should_be_processed(message, topic):
            return

        if "change.log" in topic and message is not None:
            try:
                await self._resync(message)
            except Exception as e:
                _type, _, tb = sys.exc_info()
                logger.opt(exception=(_type, None, tb)).error(
                    f"Failed to process message: {str(e)}"
                )

    async def _start(self) -> None:
        """
        The main method that starts the Kafka consumer.
        It creates a KafkaConsumer instance with the given configuration and starts it in a separate thread.
        """
        self.consumer = KafkaConsumer(
            msg_process=self._handle_message,
            config=await self._get_kafka_config(),
            org_id=self.org_id,
        )
        logger.info("Starting Kafka consumer")

        # We are running the consumer with asyncio.create_task to ensure that it runs in the background and not blocking
        # the integration's main event loop from finishing the startup process.
        self._running_task = asyncio.create_task(self.consumer.start())
        ensure_future(self._running_task)

    def _stop(self) -> None:
        if self.consumer:
            self.consumer.exit_gracefully()
        if self._running_task:
            self._running_task.cancel()



from os import environ
from typing import Tuple
import pytest

from port_ocean.clients.port.client import PortClient
from port_ocean.clients.port.types import UserAgentType
from port_ocean.tests.helpers.smoke_test import SmokeTestDetails


pytestmark = pytest.mark.smoke


@pytest.mark.skipif(
    environ.get("SMOKE_TEST_SUFFIX", None) is None,
    reason="You need to run the fake integration once",
)
async def test_valid_fake_integration(
    port_client_for_fake_integration: Tuple[SmokeTestDetails, PortClient],
) -> None:
    _, port_client = port_client_for_fake_integration
    current_integration = await port_client.get_current_integration()
    assert current_integration is not None
    assert current_integration.get("resyncState") is not None
    assert current_integration.get("resyncState", {}).get("status") == "completed"


@pytest.mark.skipif(
    environ.get("SMOKE_TEST_SUFFIX", None) is None,
    reason="You need to run the fake integration once",
)
async def test_valid_fake_departments(
    port_client_for_fake_integration: Tuple[SmokeTestDetails, PortClient],
) -> None:
    details, port_client = port_client_for_fake_integration
    entities = await port_client.search_entities(user_agent_type=UserAgentType.exporter)
    assert len(entities)
    departments = [
        x for x in entities if f"{x.blueprint}" == details.blueprint_department
    ]
    assert len(departments) == 5


@pytest.mark.skipif(
    environ.get("SMOKE_TEST_SUFFIX", None) is None,
    reason="You need to run the fake integration once",
)
async def test_valid_fake_persons(
    port_client_for_fake_integration: Tuple[SmokeTestDetails, PortClient],
) -> None:
    details, port_client = port_client_for_fake_integration
    headers = await port_client.auth.headers()
    fake_person_entities_result = await port_client.client.get(
        f"{port_client.auth.api_url}/blueprints/{details.blueprint_person}/entities",
        headers=headers,
    )

    fake_person_entities = fake_person_entities_result.json()["entities"]
    assert len(fake_person_entities)

    fake_departments_result = await port_client.client.get(
        f"{port_client.auth.api_url}/blueprints/{details.blueprint_department}/entities",
        headers=headers,
    )

    departments = [x["identifier"] for x in fake_departments_result.json()["entities"]]

    for department in departments:
        assert len(
            [
                x
                for x in fake_person_entities
                if x["relations"]["department"] == department
            ]
        )



from loguru import logger

from port_ocean.clients.port.types import UserAgentType
from port_ocean.context.event import event
from port_ocean.core.handlers.entities_state_applier.base import (
    BaseEntitiesStateApplier,
)
from port_ocean.core.handlers.entities_state_applier.port.get_related_entities import (
    get_related_entities,
)
from port_ocean.core.handlers.entities_state_applier.port.order_by_entities_dependencies import (
    order_by_entities_dependencies,
)
from port_ocean.core.models import Entity
from port_ocean.core.ocean_types import EntityDiff
from port_ocean.core.utils import is_same_entity, get_port_diff


class HttpEntitiesStateApplier(BaseEntitiesStateApplier):
    """Applies and manages changes to entities' state using HTTP requests.

    This class extends the BaseEntitiesStateApplier and provides concrete implementations
    for applying changes, deleting entities, upserting entities, and handling entity diffs
    through HTTP requests.
    """

    async def _safe_delete(
        self,
        entities_to_delete: list[Entity],
        entities_to_protect: list[Entity],
        user_agent_type: UserAgentType,
    ) -> None:
        if not entities_to_delete:
            return

        related_entities = await get_related_entities(
            entities_to_protect, self.context.port_client
        )

        allowed_entities_to_delete = []

        for entity_to_delete in entities_to_delete:
            is_part_of_related = any(
                is_same_entity(entity, entity_to_delete) for entity in related_entities
            )
            is_part_of_created = any(
                is_same_entity(entity, entity_to_delete)
                for entity in entities_to_protect
            )
            if is_part_of_related:
                if event.port_app_config.create_missing_related_entities:
                    logger.info(
                        f"Skipping entity {(entity_to_delete.identifier, entity_to_delete.blueprint)} because it is "
                        f"related to created entities and create_missing_related_entities is enabled"
                    )
                else:
                    allowed_entities_to_delete.append(entity_to_delete)
            elif not is_part_of_created:
                allowed_entities_to_delete.append(entity_to_delete)

        await self.delete(allowed_entities_to_delete, user_agent_type)

    async def apply_diff(
        self,
        entities: EntityDiff,
        user_agent_type: UserAgentType,
    ) -> None:
        diff = get_port_diff(entities["before"], entities["after"])
        kept_entities = diff.created + diff.modified

        logger.info(
            f"Updating entity diff (created: {len(diff.created)}, deleted: {len(diff.deleted)}, modified: {len(diff.modified)})"
        )
        modified_entities = await self.upsert(kept_entities, user_agent_type)

        await self._safe_delete(diff.deleted, modified_entities, user_agent_type)

    async def delete_diff(
        self,
        entities: EntityDiff,
        user_agent_type: UserAgentType,
    ) -> None:
        diff = get_port_diff(entities["before"], entities["after"])

        if not diff.deleted:
            return

        kept_entities = diff.created + diff.modified

        logger.info(
            f"Determining entities to delete ({len(diff.deleted)}/{len(kept_entities)})"
        )

        await self._safe_delete(diff.deleted, kept_entities, user_agent_type)

    async def upsert(
        self, entities: list[Entity], user_agent_type: UserAgentType
    ) -> list[Entity]:
        logger.info(f"Upserting {len(entities)} entities")
        modified_entities: list[Entity] = []
        if event.port_app_config.create_missing_related_entities:
            modified_entities = await self.context.port_client.batch_upsert_entities(
                entities,
                event.port_app_config.get_port_request_options(),
                user_agent_type,
                should_raise=False,
            )
        else:
            entities_with_search_identifier: list[Entity] = []
            entities_without_search_identifier: list[Entity] = []
            for entity in entities:
                if entity.is_using_search_identifier:
                    entities_with_search_identifier.append(entity)
                else:
                    entities_without_search_identifier.append(entity)

            ordered_created_entities = reversed(
                entities_with_search_identifier
                + order_by_entities_dependencies(entities_without_search_identifier)
            )
            for entity in ordered_created_entities:
                upsertedEntity = await self.context.port_client.upsert_entity(
                    entity,
                    event.port_app_config.get_port_request_options(),
                    user_agent_type,
                    should_raise=False,
                )
                if upsertedEntity:
                    modified_entities.append(upsertedEntity)
        return modified_entities

    async def delete(
        self, entities: list[Entity], user_agent_type: UserAgentType
    ) -> None:
        logger.info(f"Deleting {len(entities)} entities")
        if event.port_app_config.delete_dependent_entities:
            await self.context.port_client.batch_delete_entities(
                entities,
                event.port_app_config.get_port_request_options(),
                user_agent_type,
                should_raise=False,
            )
        else:
            ordered_deleted_entities = order_by_entities_dependencies(entities)

            for entity in ordered_deleted_entities:
                await self.context.port_client.delete_entity(
                    entity,
                    event.port_app_config.get_port_request_options(),
                    user_agent_type,
                    should_raise=False,
                )



# -*- coding: utf-8 -*-
from inspect import getmembers

import click

from port_ocean.bootstrap import create_default_app
from port_ocean.cli.commands.main import print_logo, console
from port_ocean.core.defaults.initialize import initialize_defaults
from port_ocean.ocean import Ocean
from port_ocean.utils.misc import load_module
from .group import defaults


@defaults.command()
@click.argument("path", default=".", type=click.Path(exists=True))
def dock(path: str) -> None:
    """
    Apply defaults of the integration from the .port/resources PATH.

    PATH: Path to the integration. If not provided, the current directory will be used.
    """
    print_logo()

    console.print("Unloading cargo at the dock... 📦🚢")

    default_app = create_default_app(path)

    main_path = f"{path}/main.py" if path else "main.py"

    app_module = load_module(main_path)
    app: Ocean = {name: item for name, item in getmembers(app_module)}.get(
        "app",
        default_app,
    )

    initialize_defaults(
        app.integration.AppConfigHandlerClass.CONFIG_CLASS,
        app.config,
    )



from typing import Union

from port_ocean.clients.port.client import PortClient


def get_port_client_for_integration(
    client_id: str,
    client_secret: str,
    integration_identifier: str,
    integration_type: str,
    integration_version: str,
    base_url: Union[str, None],
) -> PortClient:
    return PortClient(
        base_url=base_url or "https://api.getport/io",
        client_id=client_id,
        client_secret=client_secret,
        integration_identifier=integration_identifier,
        integration_type=integration_type,
        integration_version=integration_version,
    )



import sys
from inspect import getmembers, isclass
from types import ModuleType
from typing import Type, Any, Dict

from pydantic import BaseModel

from port_ocean.core.integrations.base import BaseIntegration
from port_ocean.ocean import Ocean
from port_ocean.utils.misc import load_module


def _get_base_integration_class_from_module(
    module: ModuleType,
) -> Type[BaseIntegration]:
    for name, obj in getmembers(module):
        if (
            isclass(obj)
            and type(obj) is type
            and issubclass(obj, BaseIntegration)
            and obj != BaseIntegration
        ):
            return obj

    raise Exception(f"Failed to load integration from module: {module.__name__}")


def create_default_app(
    path: str | None = None,
    config_factory: Type[BaseModel] | None = None,
    config_override: Dict[str, Any] | None = None,
) -> Ocean:
    sys.path.append(".")
    try:
        integration_path = f"{path}/integration.py" if path else "integration.py"
        module = load_module(integration_path)
        integration_class = _get_base_integration_class_from_module(module)
    except Exception:
        integration_class = None

    return Ocean(
        integration_class=integration_class,
        config_factory=config_factory,
        config_override=config_override,
    )



from contextlib import asynccontextmanager
from dataclasses import dataclass
from typing import AsyncIterator, TYPE_CHECKING

from loguru import logger
from werkzeug.local import LocalStack, LocalProxy

from port_ocean.exceptions.context import (
    ResourceContextNotFoundError,
)

if TYPE_CHECKING:
    from port_ocean.core.handlers.port_app_config.models import (
        ResourceConfig,
    )


@dataclass
class ResourceContext:
    """
    The resource context is a context manager that allows you to access the current resource config if there is one.
    This is useful for getting the resource kind / mapping from withing a resync context of a specific kind.
    """

    resource_config: "ResourceConfig"

    @property
    def kind(self) -> str:
        return self.resource_config.kind


_resource_context_stack: LocalStack[ResourceContext] = LocalStack()


def _get_resource_context() -> ResourceContext:
    """
    Get the event context from the current thread.
    """
    top_resource_context = _resource_context_stack.top
    if top_resource_context is None:
        raise ResourceContextNotFoundError(
            "You must be inside an resource context in order to use it"
        )

    return top_resource_context


resource: ResourceContext = LocalProxy(lambda: _get_resource_context())  # type: ignore


@asynccontextmanager
async def resource_context(
    resource_config: "ResourceConfig",
) -> AsyncIterator[ResourceContext]:
    _resource_context_stack.push(
        ResourceContext(
            resource_config=resource_config,
        )
    )

    with logger.contextualize(resource_kind=resource.kind):
        yield resource

    _resource_context_stack.pop()



from os import path
from typing import Any, Callable, Dict, List, Tuple, Union

import pytest
import pytest_asyncio

from port_ocean.clients.port.client import PortClient
from port_ocean.core.handlers.port_app_config.models import ResourceConfig
from port_ocean.ocean import Ocean
from port_ocean.tests.helpers.ocean_app import (
    get_integation_resource_configs,
    get_integration_ocean_app,
)
from port_ocean.tests.helpers.smoke_test import (
    SmokeTestDetails,
    get_port_client_for_fake_integration,
    get_smoke_test_details,
)


@pytest.fixture
def port_client_for_fake_integration() -> Tuple[SmokeTestDetails, PortClient]:
    smoke_test_details = get_smoke_test_details()
    port_client = get_port_client_for_fake_integration()

    return smoke_test_details, port_client


@pytest_asyncio.fixture
def get_mocked_ocean_app(request: Any) -> Callable[[], Ocean]:
    test_dir = path.join(path.dirname(request.module.__file__), "..")

    def get_ocean_app(config_overrides: Union[Dict[str, Any], None] = None) -> Ocean:
        return get_integration_ocean_app(test_dir, config_overrides)

    return get_ocean_app


@pytest_asyncio.fixture
def get_mock_ocean_resource_configs(request: Any) -> Callable[[], List[ResourceConfig]]:
    module_dir = path.join(path.dirname(request.module.__file__), "..")

    def get_ocean_resource_configs() -> List[ResourceConfig]:
        return get_integation_resource_configs(module_dir)

    return get_ocean_resource_configs



import asyncio

from loguru import logger

from port_ocean.context.event import (
    event_context,
    EventType,
)
from port_ocean.context.ocean import PortOceanContext
from port_ocean.core.event_listener.factory import (
    EventListenerFactory,
)
from port_ocean.core.integrations.mixins import SyncRawMixin, SyncMixin
from port_ocean.exceptions.core import IntegrationAlreadyStartedException


class BaseIntegration(SyncRawMixin, SyncMixin):
    """
    This is the default integration class that Ocean initializes when no custom integration class is
    provided.

    This class provides a foundation for implementing various integration types. It inherits from
    both SyncRawMixin and SyncMixin, which provide synchronization and event handling functionality.

    Parameters:
        context (PortOceanContext): The PortOceanContext object providing the necessary context
            for the integration.

    Attributes:
        started (bool): Flag indicating whether the integration has been started.
        context (PortOceanContext): The PortOceanContext object containing integration context.
        event_listener_factory (EventListenerFactory): Factory to create event listeners for
            handling integration events.

    Raises:
        IntegrationAlreadyStartedException: Raised if the integration is attempted to be started
            more than once.
        NotImplementedError: Raised if the `on_resync` method is not implemented, and the event
            strategy does not have a custom implementation for resync events.
    """

    def __init__(self, context: PortOceanContext):
        SyncRawMixin.__init__(self)
        SyncMixin.__init__(self)
        self.started = False
        self.context = context
        self.event_listener_factory = EventListenerFactory(
            context,
            self.context.config.integration.identifier,
            {"on_resync": self.sync_raw_all},
        )

    async def start(self) -> None:
        """
        Initializes handlers, establishes integration at the specified port, and starts the event listener.
        """
        logger.info("Starting integration", integration_type=self.context.config.integration.type)
        if self.started:
            raise IntegrationAlreadyStartedException("Integration already started")

        if (
            not self.event_strategy["resync"]
            and self.__class__._on_resync == BaseIntegration._on_resync
        ):
            raise NotImplementedError("on_resync is not implemented")

        await self.initialize_handlers()

        self.started = True

        async with event_context(
            EventType.START,
            trigger_type="machine",
        ):
            await asyncio.gather(
                *(listener() for listener in self.event_strategy["start"])
            )

        logger.info("Initializing event listener")
        event_listener = await self.event_listener_factory.create_event_listener()
        await event_listener.start()



import asyncio
from typing import Iterable, Any, TypeVar, Callable, Awaitable

from loguru import logger
from pydantic import parse_obj_as, ValidationError

from port_ocean.clients.port.client import PortClient
from port_ocean.core.models import Entity, Runtime
from port_ocean.core.models import EntityPortDiff
from port_ocean.core.ocean_types import RAW_RESULT
from port_ocean.exceptions.core import (
    RawObjectValidationException,
    IntegrationRuntimeException,
)

T = TypeVar("T", bound=tuple[list[Any], ...])


def zip_and_sum(collection: Iterable[T]) -> T:
    return tuple(sum(items, []) for items in zip(*collection))  # type: ignore


def validate_result(result: Any) -> RAW_RESULT:
    try:
        return parse_obj_as(list[dict[str, Any]], result)
    except ValidationError as e:
        raise RawObjectValidationException(f"Expected list[dict[str, Any]], Error: {e}")


def is_same_entity(first_entity: Entity, second_entity: Entity) -> bool:
    return (
        first_entity.identifier == second_entity.identifier
        and first_entity.blueprint == second_entity.blueprint
    )


async def validate_integration_runtime(
    port_client: PortClient,
    requested_runtime: Runtime,
) -> None:
    logger.debug("Validating integration runtime")
    current_integration = await port_client.get_current_integration(
        should_raise=False, should_log=False
    )
    current_runtime = current_integration.get("installationType", "OnPrem")
    if current_integration and current_runtime != requested_runtime.value:
        raise IntegrationRuntimeException(
            f"Invalid Runtime! Requested to run existing {current_runtime} integration in {requested_runtime} runtime."
        )


Q = TypeVar("Q")


async def gather_and_split_errors_from_results(
    task: Iterable[Awaitable[Q]],
    result_threshold_validation: Callable[[Q | Exception], bool] | None = None,
) -> tuple[list[Q], list[Exception]]:
    valid_items: list[Q] = []
    errors: list[Exception] = []
    results = await asyncio.gather(*task, return_exceptions=True)
    for item in results:
        # return_exceptions will also catch Python BaseException which also includes KeyboardInterrupt, SystemExit, GeneratorExit
        # https://docs.python.org/3/library/asyncio-task.html#asyncio.gather
        # These exceptions should be raised and not caught for the application to exit properly.
        # https://stackoverflow.com/a/17802352
        if isinstance(item, BaseException) and not isinstance(item, Exception):
            raise item
        elif isinstance(item, Exception):
            errors.append(item)
        elif not result_threshold_validation or result_threshold_validation(item):
            valid_items.append(item)

    return valid_items, errors


def get_port_diff(
    before: Iterable[Entity],
    after: Iterable[Entity],
) -> EntityPortDiff:
    before_dict = {}
    after_dict = {}
    created = []
    modified = []
    deleted = []

    # Create dictionaries for before and after lists
    for entity in before:
        key = (entity.identifier, entity.blueprint)
        before_dict[key] = entity

    for entity in after:
        key = (entity.identifier, entity.blueprint)
        after_dict[key] = entity

    # Find created, modified, and deleted objects
    for key, obj in after_dict.items():
        if key not in before_dict:
            created.append(obj)
        else:
            modified.append(obj)

    for key, obj in before_dict.items():
        if key not in after_dict:
            deleted.append(obj)

    return EntityPortDiff(created=created, modified=modified, deleted=deleted)



from jinja2 import nodes
from jinja2.ext import Extension
from jinja2.parser import Parser

from port_ocean import __version__


# Used to extend the cookiecutter syntax and allow the template to use ocean version in it using `{% version %}`
class VersionExtension(Extension):
    tags = {"version"}

    def parse(self, parser: Parser) -> nodes.Node | list[nodes.Node]:
        return nodes.Const(__version__, lineno=next(parser.stream).lineno)



from typing import Any, AsyncGenerator
import asyncio
from port_ocean.utils.async_iterators import semaphore_async_iterator
import pytest


@pytest.mark.asyncio
async def test_semaphore_async_iterator() -> None:
    max_concurrency = 5
    semaphore = asyncio.BoundedSemaphore(max_concurrency)

    concurrent_tasks = 0
    max_concurrent_tasks = 0
    lock = asyncio.Lock()  # Protect shared variables

    num_tasks = 20

    async def mock_function() -> AsyncGenerator[str, None]:
        nonlocal concurrent_tasks, max_concurrent_tasks

        async with lock:
            concurrent_tasks += 1
            if concurrent_tasks > max_concurrent_tasks:
                max_concurrent_tasks = concurrent_tasks

        await asyncio.sleep(0.1)
        yield "result"

        async with lock:
            concurrent_tasks -= 1

    async def consume_iterator(async_iterator: Any) -> None:
        async for _ in async_iterator:
            pass

    tasks = [
        consume_iterator(semaphore_async_iterator(semaphore, mock_function))
        for _ in range(num_tasks)
    ]
    await asyncio.gather(*tasks)

    assert (
        max_concurrent_tasks <= max_concurrency
    ), f"Max concurrent tasks {max_concurrent_tasks} exceeded semaphore limit {max_concurrency}"
    assert concurrent_tasks == 0, "Not all tasks have completed"



import functools
import hashlib
from typing import Callable, AsyncIterator, Any
from port_ocean.context.event import event

AsyncIteratorCallable = Callable[..., AsyncIterator[list[Any]]]


def hash_func(function_name: str, *args: Any, **kwargs: Any) -> str:
    args_str = str(args)
    kwargs_str = str(kwargs)
    concatenated_string = args_str + kwargs_str
    hash_object = hashlib.sha256(concatenated_string.encode())
    return f"{function_name}_{hash_object.hexdigest()}"


def cache_iterator_result() -> Callable[[AsyncIteratorCallable], AsyncIteratorCallable]:
    """
    This decorator caches the results of an async iterator function. It checks if the result is already in the cache
    and if not, it fetches the all the data and caches it at ocean.attributes cache the end of the iteration.

    The cache will be stored in the scope of the running event and will be removed when the event is finished.

    For example, you can use this to cache data coming back from the third-party API to avoid making the same request
    multiple times for each kind.

    The caching mechanism also detects changes in parameters.
    If a function is called with different parameter values, it will be stored in different hash keys for each unique call.

    Usage:
    ```python
    @cache_iterator_result()
    async def my_async_iterator_function():
        # Your code here
    ```
    """

    def decorator(func: AsyncIteratorCallable) -> AsyncIteratorCallable:
        @functools.wraps(func)
        async def wrapper(*args: Any, **kwargs: Any) -> Any:
            # Create Hash key from function name, args and kwargs
            cache_key = hash_func(func.__name__, *args, **kwargs)

            # Check if the result is already in the cache
            if cache := event.attributes.get(cache_key):
                yield cache
                return

            # If not in cache, fetch the data
            cached_results = list()
            async for result in func(*args, **kwargs):
                cached_results.extend(result)
                yield result

            # Cache the results
            event.attributes[cache_key] = cached_results
            return

        return wrapper

    return decorator



from contextlib import contextmanager
from typing import Awaitable, Generator, Callable

from loguru import logger

from port_ocean.core.ocean_types import (
    ASYNC_GENERATOR_RESYNC_TYPE,
    RAW_RESULT,
    RESYNC_EVENT_LISTENER,
    RESYNC_RESULT,
)
from port_ocean.core.utils import validate_result
from port_ocean.exceptions.core import (
    RawObjectValidationException,
    OceanAbortException,
    KindNotImplementedException,
)


@contextmanager
def resync_error_handling() -> Generator[None, None, None]:
    try:
        yield
    except RawObjectValidationException as error:
        raise OceanAbortException(
            f"Failed to validate raw data for returned data from resync function, error: {error}"
        ) from error
    except StopAsyncIteration:
        raise
    except Exception as error:
        err_msg = f"Failed to execute resync function, error: {error}"
        logger.exception(err_msg)
        raise OceanAbortException(err_msg) from error


async def resync_function_wrapper(
    fn: Callable[[str], Awaitable[RAW_RESULT]], kind: str
) -> RAW_RESULT:
    with resync_error_handling():
        results = await fn(kind)
        return validate_result(results)


async def resync_generator_wrapper(
    fn: Callable[[str], ASYNC_GENERATOR_RESYNC_TYPE], kind: str
) -> ASYNC_GENERATOR_RESYNC_TYPE:
    generator = fn(kind)
    errors = []
    try:
        while True:
            try:
                with resync_error_handling():
                    result = await anext(generator)
                    yield validate_result(result)
            except OceanAbortException as error:
                errors.append(error)
    except StopAsyncIteration:
        if errors:
            raise ExceptionGroup(
                "At least one of the resync generator iterations failed", errors
            )


def is_resource_supported(
    kind: str, resync_event_mapping: dict[str | None, list[RESYNC_EVENT_LISTENER]]
) -> bool:
    return bool(resync_event_mapping[kind] or resync_event_mapping[None])


def unsupported_kind_response(
    kind: str, available_resync_kinds: list[str]
) -> tuple[RESYNC_RESULT, list[Exception]]:
    logger.error(f"Kind {kind} is not supported in this integration")
    return [], [KindNotImplementedException(kind, available_resync_kinds)]



from typing import Callable, Any

from werkzeug.local import LocalProxy, LocalStack

from port_ocean.exceptions.utils import (
    SignalHandlerNotInitialized,
    SignalHandlerAlreadyInitialized,
)
from port_ocean.utils.misc import generate_uuid


class SignalHandler:
    def __init__(self) -> None:
        self._handlers: dict[str, Callable[[], Any]] = {}

    def exit(self) -> None:
        """
        Handles the exit signal.
        """
        while self._handlers:
            _, handler = self._handlers.popitem()
            handler()

    def register(self, callback: Callable[[], Any]) -> str:
        _id = generate_uuid()
        self._handlers[_id] = callback

        return _id

    def unregister(self, _id: str) -> None:
        del self._handlers[_id]


_signal_handler: LocalStack[SignalHandler] = LocalStack()


def _get_signal_handler() -> SignalHandler:
    global _signal_handler
    if _signal_handler.top is None:
        raise SignalHandlerNotInitialized("Signal handler is not initialized")
    return _signal_handler.top


signal_handler: SignalHandler = LocalProxy(_get_signal_handler)  # type: ignore


def init_signal_handler() -> None:
    global _signal_handler
    if _signal_handler.top is not None:
        raise SignalHandlerAlreadyInitialized("Signal handler is already initialized")
    _signal_handler.push(SignalHandler())



from typing import Any

import httpx
from loguru import logger

from port_ocean.clients.port.authentication import PortAuthentication
from port_ocean.clients.port.types import UserAgentType
from port_ocean.clients.port.utils import handle_status_code
from port_ocean.core.models import Blueprint


class BlueprintClientMixin:
    def __init__(self, auth: PortAuthentication, client: httpx.AsyncClient):
        self.auth = auth
        self.client = client

    async def get_blueprint(
        self, identifier: str, should_log: bool = True
    ) -> Blueprint:
        logger.info(f"Fetching blueprint with id: {identifier}")
        response = await self.client.get(
            f"{self.auth.api_url}/blueprints/{identifier}",
            headers=await self.auth.headers(),
        )
        handle_status_code(response, should_log=should_log)
        return Blueprint.parse_obj(response.json()["blueprint"])

    async def create_blueprint(
        self,
        raw_blueprint: dict[str, Any],
        user_agent_type: UserAgentType | None = None,
    ) -> dict[str, Any]:
        logger.info(f"Creating blueprint with id: {raw_blueprint.get('identifier')}")
        headers = await self.auth.headers(user_agent_type)
        response = await self.client.post(
            f"{self.auth.api_url}/blueprints", headers=headers, json=raw_blueprint
        )
        handle_status_code(response)
        return response.json()["blueprint"]

    async def patch_blueprint(
        self,
        identifier: str,
        raw_blueprint: dict[str, Any],
        user_agent_type: UserAgentType | None = None,
    ) -> None:
        logger.info(f"Patching blueprint with id: {identifier}")
        headers = await self.auth.headers(user_agent_type)
        response = await self.client.patch(
            f"{self.auth.api_url}/blueprints/{identifier}",
            headers=headers,
            json=raw_blueprint,
        )
        handle_status_code(response)

    async def delete_blueprint(
        self,
        identifier: str,
        should_raise: bool = False,
        delete_entities: bool = False,
        user_agent_type: UserAgentType | None = None,
    ) -> None | str:
        logger.info(
            f"Deleting blueprint with id: {identifier} with all entities: {delete_entities}"
        )
        headers = await self.auth.headers(user_agent_type)

        if not delete_entities:
            response = await self.client.delete(
                f"{self.auth.api_url}/blueprints/{identifier}",
                headers=headers,
            )
            handle_status_code(response, should_raise)
            return None
        else:
            response = await self.client.delete(
                f"{self.auth.api_url}/blueprints/{identifier}/all-entities?delete_blueprint=true",
                headers=await self.auth.headers(),
            )

            handle_status_code(response, should_raise)
            return response.json().get("migrationId", "")

    async def create_action(
        self, action: dict[str, Any], should_log: bool = True
    ) -> None:
        logger.info(f"Creating action: {action}")
        response = await self.client.post(
            f"{self.auth.api_url}/actions",
            json=action,
            headers=await self.auth.headers(),
        )

        handle_status_code(response, should_log=should_log)

    async def create_scorecard(
        self,
        blueprint_identifier: str,
        scorecard: dict[str, Any],
        should_log: bool = True,
    ) -> None:
        logger.info(f"Creating scorecard: {scorecard}")
        response = await self.client.post(
            f"{self.auth.api_url}/blueprints/{blueprint_identifier}/scorecards",
            json=scorecard,
            headers=await self.auth.headers(),
        )

        handle_status_code(response, should_log=should_log)

    async def create_page(
        self, page: dict[str, Any], should_log: bool = True
    ) -> dict[str, Any]:
        logger.info(f"Creating page: {page}")
        response = await self.client.post(
            f"{self.auth.api_url}/pages",
            json=page,
            headers=await self.auth.headers(),
        )

        handle_status_code(response, should_log=should_log)
        return page

    async def delete_page(
        self,
        identifier: str,
        should_raise: bool = False,
    ) -> None:
        logger.info(f"Deleting page: {identifier}")
        response = await self.client.delete(
            f"{self.auth.api_url}/pages/{identifier}",
            headers=await self.auth.headers(),
        )

        handle_status_code(response, should_raise)



import asyncio

import httpx
from loguru import logger

from port_ocean.clients.port.authentication import PortAuthentication
from port_ocean.clients.port.utils import handle_status_code
from port_ocean.core.models import Migration


class MigrationClientMixin:
    def __init__(self, auth: PortAuthentication, client: httpx.AsyncClient):
        self.auth = auth
        self.client = client

    async def wait_for_migration_to_complete(
        self,
        migration_id: str,
        interval: int = 5,
    ) -> Migration:
        logger.info(
            f"Waiting for migration with id: {migration_id} to complete",
        )

        headers = await self.auth.headers()
        response = await self.client.get(
            f"{self.auth.api_url}/migrations/{migration_id}",
            headers=headers,
        )

        handle_status_code(response, should_raise=True)

        migration_status = response.json().get("migration", {}).get("status", None)
        if (
            migration_status == "RUNNING"
            or migration_status == "INITIALIZING"
            or migration_status == "PENDING"
        ):
            await asyncio.sleep(interval)
            await self.wait_for_migration_to_complete(migration_id, interval)
        else:
            logger.info(
                f"Migration with id: {migration_id} finished with status {migration_status}",
            )

        return Migration.parse_obj(response.json()["migration"])



from typing import Any, Literal, Type, cast

from pydantic import Extra, AnyHttpUrl, parse_obj_as, parse_raw_as
from pydantic.class_validators import root_validator, validator
from pydantic.env_settings import InitSettingsSource, EnvSettingsSource, BaseSettings
from pydantic.fields import Field
from pydantic.main import BaseModel

from port_ocean.config.base import BaseOceanSettings, BaseOceanModel
from port_ocean.core.event_listener import EventListenerSettingsType
from port_ocean.core.models import Runtime
from port_ocean.utils.misc import get_integration_name, get_spec_file

LogLevelType = Literal["ERROR", "WARNING", "INFO", "DEBUG", "CRITICAL"]


class ApplicationSettings(BaseSettings):
    log_level: LogLevelType = "INFO"
    enable_http_logging: bool = True
    port: int = 8000

    class Config:
        env_prefix = "APPLICATION__"
        env_file = ".env"
        env_file_encoding = "utf-8"

        @classmethod
        def customise_sources(  # type: ignore
            cls,
            init_settings: InitSettingsSource,
            env_settings: EnvSettingsSource,
            *_,
            **__,
        ):
            return env_settings, init_settings


class PortSettings(BaseOceanModel, extra=Extra.allow):
    client_id: str = Field(..., sensitive=True)
    client_secret: str = Field(..., sensitive=True)
    base_url: AnyHttpUrl = parse_obj_as(AnyHttpUrl, "https://api.getport.io")
    port_app_config_cache_ttl: int = 60


class IntegrationSettings(BaseOceanModel, extra=Extra.allow):
    identifier: str
    type: str
    config: Any = Field(default_factory=dict)

    @root_validator(pre=True)
    def root_validator(cls, values: dict[str, Any]) -> dict[str, Any]:
        integ_type = values.get("type")

        if not integ_type:
            integ_type = get_integration_name()

        values["type"] = integ_type.lower() if integ_type else None
        if not values.get("identifier"):
            values["identifier"] = f"my-{integ_type}-integration".lower()

        return values


class IntegrationConfiguration(BaseOceanSettings, extra=Extra.allow):
    _integration_config_model: BaseModel | None = None

    allow_environment_variables_jq_access: bool = True
    initialize_port_resources: bool = True
    scheduled_resync_interval: int | None = None
    client_timeout: int = 60
    send_raw_data_examples: bool = True
    port: PortSettings
    event_listener: EventListenerSettingsType = Field(
        default=cast(EventListenerSettingsType, {"type": "POLLING"})
    )
    # If an identifier or type is not provided, it will be generated based on the integration name
    integration: IntegrationSettings = Field(
        default_factory=lambda: IntegrationSettings(type="", identifier="")
    )
    runtime: Runtime = Runtime.OnPrem

    @root_validator()
    def validate_integration_config(cls, values: dict[str, Any]) -> dict[str, Any]:
        if not (config_model := values.get("_integration_config_model")):
            return values

        # Using the integration dynamic config model to parse the config
        def parse_config(model: Type[BaseModel], config: Any) -> BaseModel:
            # In some cases, the config is parsed as a string so we need to handle it
            # Example: when the config is loaded from the environment variables and there is an object inside the config
            if isinstance(config, str):
                return parse_raw_as(model, config)
            else:
                return parse_obj_as(model, config)

        integration_config = values["integration"]
        integration_config.config = parse_config(
            config_model, integration_config.config
        )

        return values

    @validator("runtime")
    def validate_runtime(cls, runtime: Runtime) -> Runtime:
        if runtime == Runtime.Saas:
            spec = get_spec_file()
            if spec is None:
                raise ValueError(
                    "Could not determine whether it's safe to run "
                    "the integration due to not found spec.yaml."
                )

            saas_config = spec.get("saas")
            if saas_config and not saas_config["enabled"]:
                raise ValueError("This integration can't be ran as Saas")

        return runtime



import asyncio
from asyncio import Queue, Task
from typing import Any, TypeVar, Callable, Coroutine

from loguru import logger

T = TypeVar("T")


async def _start_processor_worker(
    queue: Queue[Any | None],
    func: Callable[..., Coroutine[Any, Any, T]],
    results: list[T],
    errors: list[Exception],
) -> None:
    while True:
        try:
            raw_params = await queue.get()
            if raw_params is None:
                return
            logger.debug("Processing async task")
            results.append(await func(*raw_params))
        except Exception as e:
            logger.error(f"Error processing task: {e}")
            errors.append(e)
        finally:
            queue.task_done()


async def process_in_queue(
    objects_to_process: list[Any],
    func: Callable[..., Coroutine[Any, Any, T]],
    *args: Any,
    concurrency: int = 50,
) -> list[T]:
    """
    This function executes multiple asynchronous tasks in a bounded way
    (e.g. having 200 tasks to execute, while running only 20 concurrently),
    to prevent overload and memory issues when dealing with large sets of data and tasks.
    read more -> https://stackoverflow.com/questions/38831322/making-1-milion-requests-with-aiohttp-asyncio-literally

    Usage:
    ```python
    async def incrementBy(num: int, increment_by: int) -> int:
        await asyncio.sleep(3)
        return num + increment_by

    async def main():
        raw_objects = [1, 2, 3, 4, 5]
        processed_objects = await process_in_queue(
            raw_objects,
            incrementBy,
            5
        )
    ```

    :param objects_to_process: A list of the raw objects to process
    :param func: An async function that turns raw object into result object
    :param args: Static arguments to pass to the func when called
    :param concurrency: An integer specifying the concurrent workers count
    :return: A list of result objects
    """
    queue: Queue[Any | None] = Queue(maxsize=concurrency * 2)
    tasks: list[Task[Any]] = []
    processing_results: list[T] = []
    errors: list[Exception] = []

    for i in range(concurrency):
        tasks.append(
            asyncio.create_task(
                _start_processor_worker(queue, func, processing_results, errors)
            )
        )

    for i in range(len(objects_to_process)):
        await queue.put((objects_to_process[i], *args))

    for i in range(concurrency):
        # We put None value into the queue, so the workers will know that we
        # are done sending more input and they can terminate
        await queue.put(None)

    await queue.join()
    await asyncio.gather(*tasks)
    if errors:
        raise ExceptionGroup("Error processing tasks", errors)

    return processing_results



import datetime
import signal
from typing import Literal, Any

from loguru import logger

from port_ocean.core.event_listener.base import (
    BaseEventListener,
    EventListenerEvents,
    EventListenerSettings,
)
from port_ocean.utils.repeat import repeat_every
from port_ocean.context.ocean import ocean
from port_ocean.utils.time import convert_str_to_utc_datetime, convert_to_minutes
from port_ocean.utils.misc import IntegrationStateStatus


class OnceEventListenerSettings(EventListenerSettings):
    """
    Once event listener configuration settings.
    This class inherits from `EventListenerSettings`, which provides a foundation for creating event listener settings.
    """

    type: Literal["ONCE"]

    def to_request(self) -> dict[str, Any]:
        return {}


class OnceEventListener(BaseEventListener):
    """
    Once event listener.

    This event listener is used to trigger a resync immediately.

    Parameters:
        events (EventListenerEvents): A dictionary containing event types and their corresponding event handlers.
        event_listener_config (OnceEventListenerSettings): The event listener configuration settings.
    """

    def __init__(
        self,
        events: EventListenerEvents,
        event_listener_config: OnceEventListenerSettings,
    ):
        super().__init__(events)
        self.event_listener_config = event_listener_config
        self.cached_integration: dict[str, Any] | None = None

    async def get_current_integration_cached(self) -> dict[str, Any]:
        if self.cached_integration:
            return self.cached_integration

        self.cached_integration = await ocean.port_client.get_current_integration()
        return self.cached_integration

    async def get_saas_resync_initialization_and_interval(
        self,
    ) -> tuple[int | None, datetime.datetime | None]:
        """
        Get the scheduled resync interval and the last updated time of the integration config for the saas application.
        interval is the saas configured resync interval time.
        start_time is the last updated time of the integration config.
        return: (interval, start_time)
        """
        if not ocean.app.is_saas():
            return (None, None)

        try:
            integration = await self.get_current_integration_cached()
        except Exception as e:
            logger.exception(f"Error occurred while getting current integration {e}")
            return (None, None)

        interval_str = (
            integration.get("spec", {})
            .get("appSpec", {})
            .get("scheduledResyncInterval")
        )

        if not interval_str:
            logger.error(
                "Unexpected: scheduledResyncInterval not found for Saas integration, Cannot predict the next resync"
            )
            return (None, None)

        last_updated_saas_integration_config_str = integration.get(
            "statusInfo", {}
        ).get("updatedAt")

        # we use the last updated time of the integration config as the start time since in saas application the interval is configured by the user from the portal
        if not last_updated_saas_integration_config_str:
            logger.error(
                "Unexpected: updatedAt not found for Saas integration, Cannot predict the next resync"
            )
            return (None, None)

        return (
            convert_to_minutes(interval_str),
            convert_str_to_utc_datetime(last_updated_saas_integration_config_str),
        )

    async def _before_resync(self) -> None:
        if not ocean.app.is_saas():
            # in case of non-saas, we still want to update the state before and after the resync
            await super()._before_resync()
            return

        (interval, start_time) = (
            await self.get_saas_resync_initialization_and_interval()
        )
        await ocean.app.resync_state_updater.update_before_resync(interval, start_time)

    async def _after_resync(self) -> None:
        if not ocean.app.is_saas():
            # in case of non-saas, we still want to update the state before and after the resync
            await super()._after_resync()
            return

        (interval, start_time) = (
            await self.get_saas_resync_initialization_and_interval()
        )
        await ocean.app.resync_state_updater.update_after_resync(
            IntegrationStateStatus.Completed, interval, start_time
        )

    async def _on_resync_failure(self, e: Exception) -> None:
        if not ocean.app.is_saas():
            # in case of non-saas, we still want to update the state before and after the resync
            await super()._after_resync()
            return

        (interval, start_time) = (
            await self.get_saas_resync_initialization_and_interval()
        )
        await ocean.app.resync_state_updater.update_after_resync(
            IntegrationStateStatus.Failed, interval, start_time
        )

    async def _start(self) -> None:
        """
        Starts the resync process, and exits the application once finished.
        """

        # we use the `repeat_every` decorator to make sure the resync will be triggered, but won't stuck the application
        # from finishing the startup process which is required to close the application gracefully
        @repeat_every(seconds=0, max_repetitions=1)
        async def resync_and_exit() -> None:
            logger.info("Once event listener started")
            try:
                await self._resync({})
            except Exception:
                # we catch all exceptions here to make sure the application will exit gracefully
                logger.exception("Error occurred while resyncing")
            logger.info("Once event listener finished")
            logger.info("Exiting application")
            signal.raise_signal(signal.SIGINT)

        await resync_and_exit()



from abc import abstractmethod

from port_ocean.clients.port.types import UserAgentType
from port_ocean.core.handlers.base import BaseHandler
from port_ocean.core.models import Entity
from port_ocean.core.ocean_types import EntityDiff


class BaseEntitiesStateApplier(BaseHandler):
    """Abstract base class for applying and managing changes to entities' state.

    This class defines abstract methods for applying changes, deleting entities, upserting entities,
    and deleting entity changes based on a provided user agent.

    Attributes:
        context (PortOceanContext): The context to be used during state application.
    """

    @abstractmethod
    async def apply_diff(
        self,
        entities: EntityDiff,
        user_agent: UserAgentType,
    ) -> None:
        """Apply the specified entity differences to the state.

        Args:
            entities (EntityDiff): The differences to be applied.
            user_agent (UserAgentType): The user agent responsible for the changes.
        """
        pass

    @abstractmethod
    async def delete_diff(
        self,
        entities: EntityDiff,
        user_agent: UserAgentType,
    ) -> None:
        """Delete the specified entity differences from the state.

        Args:
            entities (EntityDiff): The differences to be deleted.
            user_agent (UserAgentType): The user agent responsible for the deletion.
        """
        pass

    @abstractmethod
    async def upsert(
        self, entities: list[Entity], user_agent_type: UserAgentType
    ) -> list[Entity]:
        """Upsert (insert or update) the given entities into the state.

        Args:
            entities (list[Entity]): The entities to be upserted.
            user_agent_type (UserAgentType): The user agent responsible for the upsert.

        Returns:
            list[Entity]: The upserted entities.
        """
        pass

    @abstractmethod
    async def delete(
        self, entities: list[Entity], user_agent_type: UserAgentType
    ) -> None:
        """Delete the specified entities from the state.

        Args:
            entities (list[Entity]): The entities to be deleted.
            user_agent_type (UserAgentType): The user agent responsible for the deletion.
        """



from loguru import logger

from port_ocean.clients.port.types import UserAgentType
from port_ocean.context.ocean import ocean
from port_ocean.core.integrations.mixins.handler import HandlerMixin
from port_ocean.core.models import Entity
from port_ocean.core.ocean_types import (
    EntityDiff,
)


class SyncMixin(HandlerMixin):
    """Mixin class for synchronization of constructed entities.

    This mixin class extends the functionality of HandlerMixin to provide methods for updating,
    registering, unregistering, and syncing entities state changes.

    Note:
        Entities are constructed entities using the Entity class
    """

    def __init__(self) -> None:
        HandlerMixin.__init__(self)

    async def update_diff(
        self,
        desired_state: EntityDiff,
        user_agent_type: UserAgentType,
    ) -> None:
        """Update the state difference between two list of entities.

        - Any entities that are in the `before` state but not in the `after` state will be unregistered.
        - Any entities that are in the `after` state but not in the `before` state will be registered.
        - Any entities that are in both the `before` and `after` state will be synced.

        Args:
            desired_state (EntityDiff): The desired state difference of entities.
            user_agent_type (UserAgentType): The type of user agent.

        Raises:
            IntegrationNotStartedException: If EntitiesStateApplier class is not initialized.
        """
        await self.entities_state_applier.apply_diff(
            {"before": desired_state["before"], "after": desired_state["after"]},
            user_agent_type,
        )

    async def register(
        self,
        entities: list[Entity],
        user_agent_type: UserAgentType,
    ) -> None:
        """Upsert entities into Port.

        Args:
            entities (list[Entity]): List of entities to be registered.
            user_agent_type (UserAgentType): The type of user agent.

        Raises:
            IntegrationNotStartedException: If EntitiesStateApplier class is not initialized.
        """
        await self.entities_state_applier.upsert(entities, user_agent_type)
        logger.info("Finished registering change")

    async def unregister(
        self, entities: list[Entity], user_agent_type: UserAgentType
    ) -> None:
        """Delete entities from Port.

        Args:
            entities (list[Entity]): List of entities to be unregistered.
            user_agent_type (UserAgentType): The type of user agent.

        Raises:
            IntegrationNotStartedException: If EntitiesStateApplier class is not initialized.
        """
        await self.entities_state_applier.delete(entities, user_agent_type)
        logger.info("Finished unregistering change")

    async def sync(
        self,
        entities: list[Entity],
        user_agent_type: UserAgentType,
    ) -> None:
        """Synchronize entities' states according to the state in Port.

        The integration fetches the current state of the entities in Port according to the given user_agent_type and
        compares it to the given desired state. The integration then create/updates/delete the entities to match the
        desired state.

        Args:
            entities (list[Entity]): List of entities to be synced.
            user_agent_type (UserAgentType): The type of user agent.

        Raises:
            IntegrationNotStartedException: If EntitiesStateApplier class is not initialized.
        """
        entities_at_port = await ocean.port_client.search_entities(user_agent_type)

        modified_entities = await self.entities_state_applier.upsert(
            entities, user_agent_type
        )
        await self.entities_state_applier.delete_diff(
            {"before": entities_at_port, "after": modified_entities}, user_agent_type
        )

        logger.info("Finished syncing change")



# -*- coding: utf-8 -*-
from inspect import getmembers

import click

from port_ocean.bootstrap import create_default_app
from port_ocean.cli.commands.main import print_logo, console
from port_ocean.core.defaults import clean_defaults
from port_ocean.ocean import Ocean
from port_ocean.utils.misc import load_module
from .group import defaults


@defaults.command()
@click.argument("path", default=".", type=click.Path(exists=True))
@click.option(
    "-f",
    "--force",
    "force",
    is_flag=True,
    help="Delete all the entities of the Blueprint as well as the blueprint itself.",
)
@click.option(
    "-w",
    "--wait",
    "wait",
    is_flag=True,
    help="Wait for the migration to finish. when force is set to true.",
)
def clean(path: str, force: bool, wait: bool) -> None:
    """
    Clean defaults of the integration from the .port/resources PATH.

    PATH: Path to the integration. If not provided, the current directory will be used.
    """
    print_logo()

    console.print("Cleaning blueprints and configurations! ⚓️")

    if force:
        console.print(
            "Deleting entities forcefully I sure hope you know what you are doing 🚨 🚨 🚨 ",
        )
    default_app = create_default_app(path)

    main_path = f"{path}/main.py" if path else "main.py"

    app_module = load_module(main_path)
    app: Ocean = {name: item for name, item in getmembers(app_module)}.get(
        "app",
        default_app,
    )

    clean_defaults(app.integration.AppConfigHandlerClass.CONFIG_CLASS, force, wait)



from typing import Callable, TYPE_CHECKING, Any, Literal, Union

from fastapi import APIRouter
from pydantic.main import BaseModel
from werkzeug.local import LocalProxy

from port_ocean.clients.port.types import UserAgentType
from port_ocean.core.models import Entity
from port_ocean.core.ocean_types import (
    RESYNC_EVENT_LISTENER,
    START_EVENT_LISTENER,
    RawEntityDiff,
    EntityDiff,
)
from port_ocean.exceptions.context import (
    PortOceanContextNotFoundError,
    PortOceanContextAlreadyInitializedError,
)

if TYPE_CHECKING:
    from port_ocean.config.settings import IntegrationConfiguration
    from port_ocean.core.integrations.base import BaseIntegration
    from port_ocean.ocean import Ocean
    from port_ocean.clients.port.client import PortClient


class PortOceanContext:
    def __init__(self, app: Union["Ocean", None]) -> None:
        self._app = app

    @property
    def app(self) -> "Ocean":
        if self._app is None:
            raise PortOceanContextNotFoundError(
                "You must first initialize PortOcean in order to use it"
            )
        return self._app

    @property
    def initialized(self) -> bool:
        return self._app is not None

    @property
    def config(self) -> "IntegrationConfiguration":
        return self.app.config

    @property
    def router(self) -> APIRouter:
        return self.app.integration_router

    @property
    def integration(self) -> "BaseIntegration":
        return self.app.integration

    @property
    def integration_config(self) -> dict[str, Any]:
        if isinstance(self.app.config.integration.config, BaseModel):
            return self.app.config.integration.config.dict()
        return self.app.config.integration.config

    @property
    def port_client(self) -> "PortClient":
        return self.app.port_client

    @property
    def event_listener_type(self) -> Literal["WEBHOOK", "KAFKA", "POLLING", "ONCE"]:
        return self.app.config.event_listener.type

    def on_resync(
        self,
        kind: str | None = None,
    ) -> Callable[[RESYNC_EVENT_LISTENER], RESYNC_EVENT_LISTENER]:
        def wrapper(function: RESYNC_EVENT_LISTENER) -> RESYNC_EVENT_LISTENER:
            return self.integration.on_resync(function, kind)

        return wrapper

    def on_start(self) -> Callable[[START_EVENT_LISTENER], START_EVENT_LISTENER]:
        def wrapper(function: START_EVENT_LISTENER) -> START_EVENT_LISTENER:
            return self.integration.on_start(function)

        return wrapper

    async def update_raw_diff(
        self,
        kind: str,
        raw_diff: RawEntityDiff,
        user_agent_type: UserAgentType = UserAgentType.exporter,
    ) -> None:
        await self.integration.update_raw_diff(kind, raw_diff, user_agent_type)

    async def update_diff(
        self,
        diff: EntityDiff,
        user_agent_type: UserAgentType = UserAgentType.exporter,
    ) -> None:
        await self.integration.update_diff(diff, user_agent_type)

    async def register_raw(
        self,
        kind: str,
        change: list[dict[str, Any]],
        user_agent_type: UserAgentType = UserAgentType.exporter,
    ) -> None:
        await self.integration.register_raw(kind, change, user_agent_type)

    async def unregister_raw(
        self,
        kind: str,
        change: list[dict[str, Any]],
        user_agent_type: UserAgentType = UserAgentType.exporter,
    ) -> None:
        await self.integration.unregister_raw(kind, change, user_agent_type)

    async def register(
        self,
        entities: list[Entity],
        user_agent_type: UserAgentType = UserAgentType.exporter,
    ) -> None:
        await self.integration.register(entities, user_agent_type)

    async def unregister(
        self,
        entities: list[Entity],
        user_agent_type: UserAgentType = UserAgentType.exporter,
    ) -> None:
        await self.integration.unregister(entities, user_agent_type)

    async def sync(
        self, entities: list[Entity], user_agent_type: UserAgentType
    ) -> None:
        await self.integration.sync(entities, user_agent_type)

    async def sync_raw_all(self) -> None:
        await self.integration.sync_raw_all(trigger_type="manual")


_port_ocean: PortOceanContext = PortOceanContext(None)


def initialize_port_ocean_context(ocean_app: "Ocean") -> None:
    global _port_ocean

    if _port_ocean.initialized:
        raise PortOceanContextAlreadyInitializedError(
            "PortOcean context is already initialized"
        )
    _port_ocean = PortOceanContext(app=ocean_app)


ocean: PortOceanContext = LocalProxy(lambda: _port_ocean)  # type: ignore



import asyncio
from typing import Type, Any

import httpx
from loguru import logger

from port_ocean.clients.port.client import PortClient
from port_ocean.clients.port.types import UserAgentType
from port_ocean.config.settings import IntegrationConfiguration
from port_ocean.context.ocean import ocean
from port_ocean.core.defaults.common import (
    Defaults,
    get_port_integration_defaults,
)
from port_ocean.core.handlers.port_app_config.models import PortAppConfig
from port_ocean.core.models import Blueprint
from port_ocean.core.utils import gather_and_split_errors_from_results
from port_ocean.exceptions.port_defaults import (
    AbortDefaultCreationError,
)


def deconstruct_blueprints_to_creation_steps(
    raw_blueprints: list[dict[str, Any]],
) -> tuple[list[dict[str, Any]], ...]:
    """
    Deconstructing the blueprint into stages so the api wont fail to create a blueprint if there is a conflict
    example: Preventing the failure of creating a blueprint with a relation to another blueprint
    """
    (
        bare_blueprint,
        with_relations,
        full_blueprint,
    ) = ([], [], [])

    for blueprint in raw_blueprints.copy():
        full_blueprint.append(blueprint.copy())

        blueprint.pop("calculationProperties", {})
        blueprint.pop("mirrorProperties", {})
        blueprint.pop("aggregationProperties", {})
        with_relations.append(blueprint.copy())

        blueprint.pop("teamInheritance", {})
        blueprint.pop("relations", {})
        bare_blueprint.append(blueprint)

    return (
        bare_blueprint,
        with_relations,
        full_blueprint,
    )


async def _initialize_required_integration_settings(
    port_client: PortClient,
    default_mapping: PortAppConfig,
    integration_config: IntegrationConfiguration,
) -> None:
    try:
        logger.info("Initializing integration at port")
        integration = await port_client.get_current_integration(
            should_log=False, should_raise=False
        )
        if not integration:
            logger.info(
                "Integration does not exist, Creating new integration with default mapping"
            )
            integration = await port_client.create_integration(
                integration_config.integration.type,
                integration_config.event_listener.to_request(),
                port_app_config=default_mapping,
            )
        elif not integration.get("config"):
            logger.info(
                "Encountered that the integration's mapping is empty, Initializing to default mapping"
            )
            integration = await port_client.patch_integration(
                integration_config.integration.type,
                integration_config.event_listener.to_request(),
                port_app_config=default_mapping,
            )
    except httpx.HTTPStatusError as err:
        logger.error(f"Failed to apply default mapping: {err.response.text}.")
        raise err

    logger.info("Checking for diff in integration configuration")
    changelog_destination = integration_config.event_listener.to_request().get(
        "changelog_destination"
    )
    if (
        integration.get("changelogDestination") != changelog_destination
        or integration.get("installationAppType") != integration_config.integration.type
        or integration.get("version") != port_client.integration_version
    ):
        await port_client.patch_integration(
            integration_config.integration.type, changelog_destination
        )


async def _create_resources(
    port_client: PortClient,
    defaults: Defaults,
) -> None:
    creation_stage, *blueprint_patches = deconstruct_blueprints_to_creation_steps(
        defaults.blueprints
    )

    blueprints_results, _ = await gather_and_split_errors_from_results(
        [
            port_client.get_blueprint(blueprint["identifier"], should_log=False)
            for blueprint in creation_stage
        ],
        lambda item: isinstance(item, Blueprint),
    )

    if blueprints_results:
        logger.info(
            f"Blueprints already exist: {[result.identifier for result in blueprints_results]}. Skipping integration default creation..."
        )
        return

    created_blueprints, blueprint_errors = await gather_and_split_errors_from_results(
        (
            port_client.create_blueprint(
                blueprint, user_agent_type=UserAgentType.exporter
            )
            for blueprint in creation_stage
        )
    )

    created_blueprints_identifiers = [bp["identifier"] for bp in created_blueprints]

    if blueprint_errors:
        for error in blueprint_errors:
            if isinstance(error, httpx.HTTPStatusError):
                logger.warning(
                    f"Failed to create resources: {error.response.text}. Rolling back changes..."
                )

        raise AbortDefaultCreationError(
            created_blueprints_identifiers, blueprint_errors
        )

    try:
        for patch_stage in blueprint_patches:
            await asyncio.gather(
                *(
                    port_client.patch_blueprint(
                        blueprint["identifier"],
                        blueprint,
                        user_agent_type=UserAgentType.exporter,
                    )
                    for blueprint in patch_stage
                )
            )

    except httpx.HTTPStatusError as err:
        logger.error(f"Failed to create resources: {err.response.text}. continuing...")
        raise AbortDefaultCreationError(created_blueprints_identifiers, [err])
    try:
        created_actions, actions_errors = await gather_and_split_errors_from_results(
            (
                port_client.create_action(action, should_log=False)
                for action in defaults.actions
            )
        )

        created_scorecards, scorecards_errors = (
            await gather_and_split_errors_from_results(
                (
                    port_client.create_scorecard(
                        blueprint_scorecards["blueprint"], action, should_log=False
                    )
                    for blueprint_scorecards in defaults.scorecards
                    for action in blueprint_scorecards["data"]
                )
            )
        )

        created_pages, pages_errors = await gather_and_split_errors_from_results(
            (port_client.create_page(page, should_log=False) for page in defaults.pages)
        )

        errors = actions_errors + scorecards_errors + pages_errors
        if errors:
            for error in errors:
                if isinstance(error, httpx.HTTPStatusError):
                    logger.warning(
                        f"Failed to create resource: {error.response.text}. continuing..."
                    )

    except Exception as err:
        logger.error(f"Failed to create resources: {err}. continuing...")


async def _initialize_defaults(
    config_class: Type[PortAppConfig], integration_config: IntegrationConfiguration
) -> None:
    port_client = ocean.port_client
    defaults = get_port_integration_defaults(config_class)
    if not defaults:
        logger.warning("No defaults found. Skipping initialization...")
        return None

    if defaults.port_app_config:
        await _initialize_required_integration_settings(
            port_client, defaults.port_app_config, integration_config
        )

    if not integration_config.initialize_port_resources:
        return

    try:
        logger.info("Found default resources, starting creation process")
        await _create_resources(port_client, defaults)
    except AbortDefaultCreationError as e:
        logger.warning(
            f"Failed to create resources. Rolling back blueprints : {e.blueprints_to_rollback}"
        )
        await asyncio.gather(
            *(
                port_client.delete_blueprint(
                    identifier,
                    should_raise=False,
                    user_agent_type=UserAgentType.exporter,
                )
                for identifier in e.blueprints_to_rollback
            )
        )
        raise ExceptionGroup(str(e), e.errors)


def initialize_defaults(
    config_class: Type[PortAppConfig], integration_config: IntegrationConfiguration
) -> None:
    asyncio.new_event_loop().run_until_complete(
        _initialize_defaults(config_class, integration_config)
    )



from typing import Any
from unittest.mock import AsyncMock, Mock
import pytest

from port_ocean.context.ocean import PortOceanContext
from port_ocean.core.handlers.entity_processor.jq_entity_processor import (
    JQEntityProcessor,
)
from port_ocean.core.ocean_types import CalculationResult
from port_ocean.exceptions.core import EntityProcessorException


@pytest.mark.asyncio
class TestJQEntityProcessor:

    @pytest.fixture
    def mocked_processor(self, monkeypatch: Any) -> JQEntityProcessor:
        mock_context = AsyncMock()
        monkeypatch.setattr(PortOceanContext, "app", mock_context)
        return JQEntityProcessor(mock_context)

    async def test_compile(self, mocked_processor: JQEntityProcessor) -> None:
        pattern = ".foo"
        compiled = mocked_processor._compile(pattern)
        assert compiled is not None

    async def test_search(self, mocked_processor: JQEntityProcessor) -> None:
        data = {"foo": "bar"}
        pattern = ".foo"
        result = await mocked_processor._search(data, pattern)
        assert result == "bar"

    async def test_search_as_bool(self, mocked_processor: JQEntityProcessor) -> None:
        data = {"foo": True}
        pattern = ".foo"
        result = await mocked_processor._search_as_bool(data, pattern)
        assert result is True

    async def test_search_as_object(self, mocked_processor: JQEntityProcessor) -> None:
        data = {"foo": {"bar": "baz"}}
        obj = {"foo": ".foo.bar"}
        result = await mocked_processor._search_as_object(data, obj)
        assert result == {"foo": "baz"}

    async def test_get_mapped_entity(self, mocked_processor: JQEntityProcessor) -> None:
        data = {"foo": "bar"}
        raw_entity_mappings = {"foo": ".foo"}
        selector_query = '.foo == "bar"'
        result = await mocked_processor._get_mapped_entity(
            data, raw_entity_mappings, selector_query
        )
        assert result.entity == {"foo": "bar"}
        assert result.did_entity_pass_selector is True

    async def test_calculate_entity(self, mocked_processor: JQEntityProcessor) -> None:
        data = {"foo": "bar"}
        raw_entity_mappings = {"foo": ".foo"}
        selector_query = '.foo == "bar"'
        result, errors = await mocked_processor._calculate_entity(
            data, raw_entity_mappings, None, selector_query
        )
        assert len(result) == 1
        assert result[0].entity == {"foo": "bar"}
        assert result[0].did_entity_pass_selector is True
        assert not errors

    async def test_parse_items(self, mocked_processor: JQEntityProcessor) -> None:
        mapping = Mock()
        mapping.port.entity.mappings.dict.return_value = {
            "identifier": ".foo",
            "blueprint": ".foo",
            "properties": {"foo": ".foo"},
        }
        mapping.port.items_to_parse = None
        mapping.selector.query = '.foo == "bar"'
        raw_results = [{"foo": "bar"}]
        result = await mocked_processor._parse_items(mapping, raw_results)
        assert isinstance(result, CalculationResult)
        assert len(result.entity_selector_diff.passed) == 1
        assert result.entity_selector_diff.passed[0].properties.get("foo") == "bar"
        assert not result.errors

    async def test_in_operator(self, mocked_processor: JQEntityProcessor) -> None:
        data = {
            "key": "GetPort_SelfService",
            "name": "GetPort SelfService",
            "desc": "Test",
            "qualifier": "VW",
            "visibility": "public",
            "selectionMode": "NONE",
            "subViews": [
                {
                    "key": "GetPort_SelfService_Second",
                    "name": "GetPort SelfService Second",
                    "qualifier": "SVW",
                    "selectionMode": "NONE",
                    "subViews": [
                        {
                            "key": "GetPort_SelfService_Third",
                            "name": "GetPort SelfService Third",
                            "qualifier": "SVW",
                            "selectionMode": "NONE",
                            "subViews": [],
                            "referencedBy": [],
                        },
                        {
                            "key": "Port_Test",
                            "name": "Port Test",
                            "qualifier": "SVW",
                            "selectionMode": "NONE",
                            "subViews": [],
                            "referencedBy": [],
                        },
                    ],
                    "referencedBy": [],
                },
                {
                    "key": "Python",
                    "name": "Python",
                    "qualifier": "SVW",
                    "selectionMode": "NONE",
                    "subViews": [
                        {
                            "key": "Time",
                            "name": "Time",
                            "qualifier": "SVW",
                            "selectionMode": "NONE",
                            "subViews": [
                                {
                                    "key": "port_*****",
                                    "name": "port-*****",
                                    "qualifier": "SVW",
                                    "selectionMode": "NONE",
                                    "subViews": [
                                        {
                                            "key": "port_*****:REferenced",
                                            "name": "REferenced",
                                            "qualifier": "VW",
                                            "visibility": "public",
                                            "originalKey": "REferenced",
                                        }
                                    ],
                                    "referencedBy": [],
                                }
                            ],
                            "referencedBy": [],
                        }
                    ],
                    "referencedBy": [],
                },
                {
                    "key": "GetPort_SelfService:Authentication_Application",
                    "name": "Authentication Application",
                    "desc": "For auth services",
                    "qualifier": "APP",
                    "visibility": "private",
                    "selectedBranches": ["main"],
                    "originalKey": "Authentication_Application",
                },
            ],
            "referencedBy": [],
        }
        pattern = '.subViews | map(select((.qualifier | IN("VW", "SVW"))) | .key)'
        result = await mocked_processor._search(data, pattern)
        assert result == ["GetPort_SelfService_Second", "Python"]

    async def test_failure_of_jq_expression(
        self, mocked_processor: JQEntityProcessor
    ) -> None:
        data = {"foo": "bar"}
        pattern = ".foo."
        result = await mocked_processor._search(data, pattern)
        assert result is None

    async def test_search_as_object_failure(
        self, mocked_processor: JQEntityProcessor
    ) -> None:
        data = {"foo": {"bar": "baz"}}
        obj = {"foo": ".foo.bar."}
        result = await mocked_processor._search_as_object(data, obj)
        assert result == {"foo": None}

    async def test_double_quotes_in_jq_expression(
        self, mocked_processor: JQEntityProcessor
    ) -> None:
        data = {"foo": "bar"}
        pattern = '"shalom"'
        result = await mocked_processor._search(data, pattern)
        assert result == "shalom"

    async def test_search_as_bool_failure(
        self, mocked_processor: JQEntityProcessor
    ) -> None:
        data = {"foo": "bar"}
        pattern = ".foo"
        with pytest.raises(
            EntityProcessorException,
            match="Expected boolean value, got value:bar of type: <class 'str'> instead",
        ):
            await mocked_processor._search_as_bool(data, pattern)

    @pytest.mark.parametrize(
        "pattern, expected",
        [
            ('.parameters[] | select(.name == "not_exists") | .value', None),
            (
                '.parameters[] | select(.name == "parameter_name") | .value',
                "parameter_value",
            ),
            (
                '.parameters[] | select(.name == "another_parameter") | .value',
                "another_value",
            ),
        ],
    )
    async def test_search_fails_on_stop_iteration(
        self, mocked_processor: JQEntityProcessor, pattern: str, expected: Any
    ) -> None:
        data = {
            "parameters": [
                {"name": "parameter_name", "value": "parameter_value"},
                {"name": "another_parameter", "value": "another_value"},
                {"name": "another_parameter", "value": "another_value2"},
            ]
        }
        result = await mocked_processor._search(data, pattern)
        assert result == expected

    async def test_return_a_list_of_values(
        self, mocked_processor: JQEntityProcessor
    ) -> None:
        data = {"parameters": ["parameter_value", "another_value", "another_value2"]}
        pattern = ".parameters"
        result = await mocked_processor._search(data, pattern)
        assert result == ["parameter_value", "another_value", "another_value2"]

    @pytest.mark.timeout(3)
    async def test_search_performance_10000(
        self, mocked_processor: JQEntityProcessor
    ) -> None:
        """
        This test is to check the performance of the search method when called 10000 times.
        """
        data = {"foo": "bar"}
        pattern = ".foo"
        for _ in range(10000):
            result = await mocked_processor._search(data, pattern)
            assert result == "bar"

    @pytest.mark.timeout(15)
    async def test_parse_items_performance_10000(
        self, mocked_processor: JQEntityProcessor
    ) -> None:
        """
        This test is to check the performance of the parse_items method when called 10000 times.
        """
        mapping = Mock()
        mapping.port.entity.mappings.dict.return_value = {
            "identifier": ".foo",
            "blueprint": ".foo",
            "properties": {"foo": ".foo"},
        }
        mapping.port.items_to_parse = None
        mapping.selector.query = '.foo == "bar"'
        raw_results = [{"foo": "bar"}]
        for _ in range(10000):
            result = await mocked_processor._parse_items(mapping, raw_results)
            assert isinstance(result, CalculationResult)
            assert len(result.entity_selector_diff.passed) == 1
            assert result.entity_selector_diff.passed[0].properties.get("foo") == "bar"
            assert not result.errors



from port_ocean.exceptions.base import BaseOceanException


class NoContextError(BaseOceanException):
    pass


class ResourceContextNotFoundError(NoContextError):
    pass


class EventContextNotFoundError(NoContextError):
    pass


class PortOceanContextNotFoundError(NoContextError):
    pass


class PortOceanContextAlreadyInitializedError(NoContextError):
    pass



import warnings


warnings.filterwarnings("ignore", category=FutureWarning)

from .ocean import Ocean  # noqa: E402
from .run import run  # noqa: E402
from .version import __integration_version__, __version__  # noqa: E402


__all__ = ["Ocean", "run", "__version__", "__integration_version__"]



import asyncio
from typing import Type

import httpx
from loguru import logger

from port_ocean.context.ocean import ocean
from port_ocean.core.defaults.common import (
    get_port_integration_defaults,
    is_integration_exists,
)
from port_ocean.core.handlers.port_app_config.models import PortAppConfig


def clean_defaults(
    config_class: Type[PortAppConfig],
    force: bool,
    wait: bool,
) -> None:
    try:
        asyncio.new_event_loop().run_until_complete(
            _clean_defaults(config_class, force, wait)
        )

    except Exception as e:
        logger.error(f"Failed to clear defaults, skipping... Error: {e}")


async def _clean_defaults(
    config_class: Type[PortAppConfig], force: bool, wait: bool
) -> None:
    port_client = ocean.port_client
    is_exists = await is_integration_exists(port_client)
    if not is_exists:
        return None
    defaults = get_port_integration_defaults(config_class)
    if not defaults:
        return None

    try:
        delete_result = await asyncio.gather(
            *(
                port_client.delete_blueprint(
                    blueprint["identifier"], should_raise=True, delete_entities=force
                )
                for blueprint in defaults.blueprints
            )
        )

        if not force:
            logger.info(
                "Finished deleting blueprints and configurations! ⚓️",
            )
            return None

        migration_ids = [migration_id for migration_id in delete_result if migration_id]

        if migration_ids and len(migration_ids) > 0 and not wait:
            logger.info(
                f"Migration started. To check the status of the migration, track these ids using /migrations/:id route {migration_ids}",
            )
        elif migration_ids and len(migration_ids) > 0 and wait:
            await asyncio.gather(
                *(
                    ocean.port_client.wait_for_migration_to_complete(migration_id)
                    for migration_id in migration_ids
                )
            )
    except httpx.HTTPStatusError as e:
        logger.error(f"Failed to delete blueprints: {e.response.text}.")
        raise e



from typing import List

from loguru import logger

from port_ocean.clients.port.client import PortClient


async def cleanup_integration(client: PortClient, blueprints: List[str]) -> None:
    for blueprint in blueprints:
        try:
            bp = await client.get_blueprint(blueprint)
            if bp is not None:
                migration_id = await client.delete_blueprint(
                    identifier=blueprint, delete_entities=True
                )
                if migration_id:
                    await client.wait_for_migration_to_complete(
                        migration_id=migration_id
                    )
        except Exception as bp_e:
            logger.info(f"Skipping missing blueprint ({blueprint}): {bp_e}")
    headers = await client.auth.headers()
    try:
        await client.client.delete(
            f"{client.auth.api_url}/integrations/{client.integration_identifier}",
            headers=headers,
        )
    except Exception as int_e:
        logger.info(
            f"Failed to delete integration ({client.integration_identifier}): {int_e}"
        )



from typing import Any, TYPE_CHECKING, Optional, TypedDict
from urllib.parse import quote_plus

import httpx
from loguru import logger
from port_ocean.clients.port.authentication import PortAuthentication
from port_ocean.clients.port.utils import handle_status_code
from port_ocean.log.sensetive import sensitive_log_filter

if TYPE_CHECKING:
    from port_ocean.core.handlers.port_app_config.models import PortAppConfig


class LogAttributes(TypedDict):
    ingestUrl: str


class IntegrationClientMixin:
    def __init__(
        self,
        integration_identifier: str,
        integration_version: str,
        auth: PortAuthentication,
        client: httpx.AsyncClient,
    ):
        self.integration_identifier = integration_identifier
        self.integration_version = integration_version
        self.auth = auth
        self.client = client
        self._log_attributes: LogAttributes | None = None

    async def _get_current_integration(self) -> httpx.Response:
        logger.info(f"Fetching integration with id: {self.integration_identifier}")
        response = await self.client.get(
            f"{self.auth.api_url}/integration/{self.integration_identifier}",
            headers=await self.auth.headers(),
        )
        return response

    async def get_current_integration(
        self, should_raise: bool = True, should_log: bool = True
    ) -> dict[str, Any]:
        response = await self._get_current_integration()
        handle_status_code(response, should_raise, should_log)
        return response.json().get("integration", {})

    async def get_log_attributes(self) -> LogAttributes:
        if self._log_attributes is None:
            response = await self.get_current_integration()
            self._log_attributes = response["logAttributes"]
        return self._log_attributes

    async def create_integration(
        self,
        _type: str,
        changelog_destination: dict[str, Any],
        port_app_config: Optional["PortAppConfig"] = None,
    ) -> dict:
        logger.info(f"Creating integration with id: {self.integration_identifier}")
        headers = await self.auth.headers()
        json = {
            "installationId": self.integration_identifier,
            "installationAppType": _type,
            "version": self.integration_version,
            "changelogDestination": changelog_destination,
            "config": {},
        }
        if port_app_config:
            json["config"] = port_app_config.to_request()
        response = await self.client.post(
            f"{self.auth.api_url}/integration", headers=headers, json=json
        )
        handle_status_code(response)
        return response.json()["integration"]

    async def patch_integration(
        self,
        _type: str | None = None,
        changelog_destination: dict[str, Any] | None = None,
        port_app_config: Optional["PortAppConfig"] = None,
    ) -> dict:
        logger.info(f"Updating integration with id: {self.integration_identifier}")
        headers = await self.auth.headers()
        json: dict[str, Any] = {}
        if _type:
            json["installationAppType"] = _type
        if changelog_destination:
            json["changelogDestination"] = changelog_destination
        if port_app_config:
            json["config"] = port_app_config.to_request()
        json["version"] = self.integration_version

        response = await self.client.patch(
            f"{self.auth.api_url}/integration/{self.integration_identifier}",
            headers=headers,
            json=json,
        )
        handle_status_code(response)
        return response.json()["integration"]

    async def ingest_integration_logs(self, logs: list[dict[str, Any]]) -> None:
        logger.debug("Ingesting logs")
        log_attributes = await self.get_log_attributes()
        headers = await self.auth.headers()
        response = await self.client.post(
            log_attributes["ingestUrl"],
            headers=headers,
            json={
                "logs": logs,
            },
        )
        handle_status_code(response, should_log=False)
        logger.debug("Logs successfully ingested")

    async def ingest_integration_kind_examples(
        self, kind: str, data: list[dict[str, Any]], should_log: bool = True
    ):
        logger.debug(f"Ingesting examples for kind: {kind}")
        headers = await self.auth.headers()
        response = await self.client.post(
            f"{self.auth.api_url}/integration/{quote_plus(self.integration_identifier)}/kinds/{quote_plus(kind)}/examples",
            headers=headers,
            json={
                "examples": sensitive_log_filter.mask_object(data, full_hide=True),
            },
        )
        handle_status_code(response, should_log=should_log)
        logger.debug(f"Examples for kind {kind} successfully ingested")



import functools
import signal
from asyncio import get_running_loop, ensure_future
from typing import Any, Callable, Awaitable

from confluent_kafka import Consumer, KafkaException, Message  # type: ignore
from loguru import logger
from pydantic import BaseModel


class KafkaConsumerConfig(BaseModel):
    brokers: str
    username: str | None = None
    password: str | None = None
    group_name: str | None = None
    security_protocol: str
    authentication_mechanism: str
    kafka_security_enabled: bool
    consumer_poll_timeout: int


class KafkaConsumer:
    def __init__(
        self,
        msg_process: Callable[[Message], Awaitable[None]],
        config: KafkaConsumerConfig,
        org_id: str,
    ) -> None:
        self.running = False
        self._assigned_partitions = False
        self.org_id = org_id
        self.config = config

        self.msg_process = msg_process
        if config.kafka_security_enabled:
            kafka_config = {
                "bootstrap.servers": config.brokers,
                "security.protocol": config.security_protocol,
                "sasl.mechanism": config.authentication_mechanism,
                "sasl.username": config.username,
                "sasl.password": config.password,
                "group.id": f"{self.org_id}.{config.group_name}",
                "enable.auto.commit": "false",
                # We use the cooperative-sticky assignment strategy to ensure that the same instance of the integration
                # is always assigned the same partitions. This is important as only one instance of the integration
                # can be assigned to a partition at a time and we dont want a running instance to lose its partitions
                # when a new instance starts and causes a rebalance.
                "partition.assignment.strategy": "cooperative-sticky",
            }
        else:
            kafka_config = {
                "bootstrap.servers": config.brokers,
                "group.id": "no-security",
                "enable.auto.commit": "false",
            }

        self.consumer = Consumer(kafka_config)

    def _handle_partitions_assignment(self, _: Any, partitions: list[str]) -> None:
        logger.info(f"Assigned partitions: {partitions}")
        if not partitions and not self._assigned_partitions:
            logger.error(
                "No partitions assigned. This usually means that there is"
                " already another integration from the same type and with"
                " the same identifier running. Two integrations of the same"
                " type and identifier cannot run at the same time."
            )
            signal.raise_signal(signal.SIGINT)
        else:
            self._assigned_partitions = True

    async def start(self) -> None:
        self.running = True
        logger.info("Starting kafka consumer...")
        topics = [f"{self.org_id}.change.log"]
        self.consumer.subscribe(
            topics,
            on_assign=self._handle_partitions_assignment,
        )
        logger.info(f"Subscribed to topics: {topics}")

        loop = get_running_loop()
        poll = functools.partial(
            self.consumer.poll, timeout=self.config.consumer_poll_timeout
        )
        try:
            while self.running:
                try:
                    msg = await loop.run_in_executor(None, poll)
                    if msg is None:
                        continue
                    if msg.error():
                        raise KafkaException(msg.error())
                    else:
                        try:
                            logger.info(
                                "Process message "
                                f"from topic {msg.topic()}, partition {msg.partition()}, offset {msg.offset()}"
                            )
                            ensure_future(self.msg_process(msg))

                        except Exception as process_error:
                            logger.exception(
                                "Failed process message"
                                f" from topic {msg.topic()}, partition {msg.partition()}, offset {msg.offset()}: {str(process_error)}"
                            )
                        finally:
                            self.consumer.commit(asynchronous=False)
                except Exception as message_error:
                    logger.error(str(message_error))
        finally:
            logger.info("Closing consumer...")
            self.exit_gracefully()

    def exit_gracefully(self, *_: Any) -> None:
        logger.info("Closing the kafka consumer gracefully...")
        self.running = False
        self.consumer.close()



from port_ocean.exceptions.base import BaseOceanException


class OceanAbortException(BaseOceanException):
    pass


class KindNotImplementedException(OceanAbortException):
    def __init__(self, kind: str, available_kinds: list[str]):
        base_message = f"Kind {kind} is not implemented."
        super().__init__(f"{base_message} Available kinds: {available_kinds}")


class RawObjectValidationException(OceanAbortException):
    pass


class EntityProcessorException(BaseOceanException):
    pass


class RelationValidationException(OceanAbortException):
    pass


class UnsupportedEventListenerTypeException(BaseOceanException):
    pass


class IntegrationAlreadyStartedException(BaseOceanException):
    pass


class IntegrationNotStartedException(BaseOceanException):
    pass


class IntegrationRuntimeException(BaseOceanException):
    pass



import asyncio
import sys
import threading
from contextlib import asynccontextmanager
from typing import Callable, Any, Dict, AsyncIterator, Type

from fastapi import FastAPI, APIRouter
from loguru import logger
from pydantic import BaseModel
from starlette.types import Scope, Receive, Send

from port_ocean.core.handlers.resync_state_updater import ResyncStateUpdater
from port_ocean.core.models import Runtime
from port_ocean.clients.port.client import PortClient
from port_ocean.config.settings import (
    IntegrationConfiguration,
)
from port_ocean.context.ocean import (
    PortOceanContext,
    ocean,
    initialize_port_ocean_context,
)
from port_ocean.core.integrations.base import BaseIntegration
from port_ocean.log.sensetive import sensitive_log_filter
from port_ocean.middlewares import request_handler
from port_ocean.utils.repeat import repeat_every
from port_ocean.utils.signal import signal_handler
from port_ocean.version import __integration_version__
from port_ocean.utils.misc import IntegrationStateStatus


class Ocean:
    def __init__(
        self,
        app: FastAPI | None = None,
        integration_class: Callable[[PortOceanContext], BaseIntegration] | None = None,
        integration_router: APIRouter | None = None,
        config_factory: Type[BaseModel] | None = None,
        config_override: Dict[str, Any] | None = None,
    ):
        initialize_port_ocean_context(self)

        @asynccontextmanager
        async def lifespan(_: FastAPI) -> AsyncIterator[None]:
            try:
                await self.integration.start()
                await self._setup_scheduled_resync()
                yield None
            except Exception:
                logger.exception("Integration had a fatal error. Shutting down.")
                logger.complete()
                sys.exit("Server stopped")
            finally:
                signal_handler.exit()

        self.fast_api_app = app or FastAPI(lifespan=lifespan)
        self.fast_api_app.middleware("http")(request_handler)

        self.config = IntegrationConfiguration(
            # type: ignore
            _integration_config_model=config_factory,
            **(config_override or {}),
        )

        # add the integration sensitive configuration to the sensitive patterns to mask out
        sensitive_log_filter.hide_sensitive_strings(
            *self.config.get_sensitive_fields_data()
        )
        self.integration_router = integration_router or APIRouter()
        self.integration_router_included = False

        self.port_client = PortClient(
            base_url=self.config.port.base_url,
            client_id=self.config.port.client_id,
            client_secret=self.config.port.client_secret,
            integration_identifier=self.config.integration.identifier,
            integration_type=self.config.integration.type,
            integration_version=__integration_version__,
        )
        self.integration = (
            integration_class(ocean) if integration_class else BaseIntegration(ocean)
        )

        self.resync_state_updater = ResyncStateUpdater(
            self.port_client, self.config.scheduled_resync_interval
        )

    def is_saas(self) -> bool:
        return self.config.runtime == Runtime.Saas

    async def _setup_scheduled_resync(
        self,
    ) -> None:
        async def execute_resync_all() -> None:
            await self.resync_state_updater.update_before_resync()
            logger.info("Starting a new scheduled resync")
            try:
                await self.integration.sync_raw_all()
                await self.resync_state_updater.update_after_resync()
            except asyncio.CancelledError:
                logger.warning(
                    "resync was cancelled by the scheduled resync, skipping state update"
                )
            except Exception as e:
                await self.resync_state_updater.update_after_resync(
                    IntegrationStateStatus.Failed
                )
                raise e

        interval = self.config.scheduled_resync_interval
        loop = asyncio.get_event_loop()
        if interval is not None:
            logger.info(
                f"Setting up scheduled resync, the integration will automatically perform a full resync every {interval} minutes)",
                scheduled_interval=interval,
            )
            repeated_function = repeat_every(
                seconds=interval * 60,
                # Not running the resync immediately because the event listener should run resync on startup
                wait_first=True,
            )(
                lambda: threading.Thread(
                    target=lambda: asyncio.run_coroutine_threadsafe(
                        execute_resync_all(), loop
                    )
                ).start()
            )
            await repeated_function()

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if not self.integration_router_included:
            self.fast_api_app.include_router(
                self.integration_router, prefix="/integration"
            )
            self.integration_router_included = True

        await self.fast_api_app(scope, receive, send)



import asyncio
from http import HTTPStatus
from typing import TYPE_CHECKING, Any

import httpx

from port_ocean.helpers.retry import RetryTransport

if TYPE_CHECKING:
    from port_ocean.clients.port.client import PortClient


class TokenRetryTransport(RetryTransport):
    def __init__(self, port_client: "PortClient", **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.port_client = port_client

    async def _handle_unauthorized(self, response: httpx.Response) -> None:
        token = await self.port_client.auth.token
        response.headers["Authorization"] = f"Bearer {token}"

    def is_token_error(self, response: httpx.Response) -> bool:
        return (
            response.status_code == HTTPStatus.UNAUTHORIZED
            and "/auth/access_token" not in str(response.request.url)
            and self.port_client.auth.last_token_object is not None
            and self.port_client.auth.last_token_object.expired
        )

    async def _should_retry_async(self, response: httpx.Response) -> bool:
        if self.is_token_error(response):
            if self._logger:
                self._logger.info(
                    "Got unauthorized response, trying to refresh token before retrying"
                )
            await self._handle_unauthorized(response)
            return True
        return await super()._should_retry_async(response)

    def _should_retry(self, response: httpx.Response) -> bool:
        if self.is_token_error(response):
            if self._logger:
                self._logger.info(
                    "Got unauthorized response, trying to refresh token before retrying"
                )
            asyncio.get_running_loop().run_until_complete(
                self._handle_unauthorized(response)
            )

            return True
        return super()._should_retry(response)



from typing import Any

from loguru import logger

from port_ocean.core.handlers.port_app_config.base import BasePortAppConfig


class APIPortAppConfig(BasePortAppConfig):
    """Class for obtaining port application configuration through an API.

    This class extends the BasePortAppConfig and provides a method to fetch
    port application configuration from an API integration.
    """

    async def _get_port_app_config(self) -> dict[str, Any]:
        logger.info("Fetching port app config")
        integration = await self.context.port_client.get_current_integration()
        config = integration["config"]

        if not config:
            logger.error(
                "The integration port app config is empty. "
                "Please make sure to configure your port app config using Port's API."
            )

        return config



import os
import re
from types import GenericAlias
from typing import Any

import yaml
from humps import decamelize
from pathlib import Path
from pydantic import BaseSettings
from pydantic.env_settings import EnvSettingsSource, InitSettingsSource
from pydantic.main import ModelMetaclass, BaseModel

PROVIDER_WRAPPER_PATTERN = r"{{ from (.*) }}"
PROVIDER_CONFIG_PATTERN = r"^[a-zA-Z0-9]+ .*$"


def read_yaml_config_settings_source(settings: "BaseOceanSettings") -> dict[str, Any]:
    yaml_file = getattr(settings.Config, "yaml_file", "")

    assert yaml_file, "Settings.yaml_file not properly configured"
    path = Path(
        getattr(
            settings,
            "_base_path",
        ),
        yaml_file,
    )

    if not path.exists():
        return {}
    return yaml.safe_load(path.read_text("utf-8"))


def parse_config_provider(value: str) -> tuple[str, str]:
    match = re.match(PROVIDER_CONFIG_PATTERN, value)
    if not match:
        raise ValueError(
            f"Invalid pattern: {value}. Pattern should match: {PROVIDER_CONFIG_PATTERN}"
        )

    provider_type, provider_value = value.split(" ", 1)

    return provider_type, provider_value


def load_from_config_provider(config_provider: str) -> Any:
    provider_type, value = parse_config_provider(config_provider)
    if provider_type == "env":
        result = os.environ.get(value)
        if result is None:
            raise ValueError(f"Environment variable not found: {value}")
        return result
    else:
        raise ValueError(f"Invalid provider type: {provider_type}")


def parse_providers(
    settings_model: BaseModel | ModelMetaclass,
    config: dict[str, Any],
    existing_data: dict[str, Any],
) -> dict[str, Any]:
    """
    Normalizing the config yaml file to work with snake_case and getting only the data that is missing for the settings
    """
    for key, value in config.items():
        if isinstance(value, dict) and settings_model is not None:
            # If the value is of type ModelMetaClass then its a nested model, and we need to parse it
            # If the value is of type primitive dict then we need to decamelize the keys and not recurse into the values because its no longer part of the model
            _type = settings_model.__annotations__[key]
            is_primitive_dict_type = _type is dict or (
                isinstance(_type, GenericAlias) and _type.__origin__ is dict
            )

            if is_primitive_dict_type:
                _type = None
            existing_data[key] = parse_providers(
                _type, value, existing_data.get(key, {})
            )

        elif isinstance(value, str):
            # If the value is a provider, we try to load it from the provider
            if provider_match := re.match(PROVIDER_WRAPPER_PATTERN, value):
                # If the there is already value for that field, we ignore it
                # If the provider failed to load, we ignore it
                if key not in existing_data:
                    try:
                        existing_data[key] = load_from_config_provider(
                            provider_match.group(1)
                        )
                    except ValueError:
                        pass
            else:
                existing_data[key] = value
        else:
            existing_data[key] = value
    return existing_data


def decamelize_config(
    settings_model: BaseModel | ModelMetaclass, config: dict[str, Any]
) -> dict[str, Any]:
    """
    Normalizing the config yaml file to work with snake_case and getting only the data that is missing for the settings
    """
    result = {}
    for key, value in config.items():
        decamelize_key = decamelize(key)
        if isinstance(value, dict) and settings_model is not None:
            # If the value is ModelMetaClass typed then its a nested model, and we need to parse it
            # If the value is a primitive dict then we need to decamelize the keys and not recurse into the values because its no longer part of the model
            _type = settings_model.__annotations__[decamelize_key]
            is_primitive_dict_type = _type is dict or (
                isinstance(_type, GenericAlias) and _type.__origin__ is dict
            )

            if is_primitive_dict_type:
                _type = None

            result[decamelize_key] = decamelize_config(_type, value)
        else:
            result[decamelize_key] = value
    return result


def load_providers(
    settings: "BaseOceanSettings", existing_values: dict[str, Any]
) -> dict[str, Any]:
    data = read_yaml_config_settings_source(settings)
    snake_case_config = decamelize_config(settings, data)
    return parse_providers(settings, snake_case_config, existing_values)


class BaseOceanSettings(BaseSettings):
    _base_path: str = "./"

    def get_sensitive_fields_data(self) -> set[str]:
        return _get_sensitive_information(self)

    class Config:
        yaml_file = "./config.yaml"
        env_prefix = "OCEAN__"
        env_nested_delimiter = "__"
        env_file = ".env"
        env_file_encoding = "utf-8"

        @classmethod
        def customise_sources(  # type: ignore
            cls,
            init_settings: InitSettingsSource,
            env_settings: EnvSettingsSource,
            *_,
            **__,
        ):
            return (
                init_settings,
                env_settings,
                lambda s: load_providers(s, {**env_settings(s), **init_settings(s)}),
            )


class BaseOceanModel(BaseModel):
    def get_sensitive_fields_data(self) -> set[str]:
        return _get_sensitive_information(self)


def _get_sensitive_information(
    model: BaseOceanModel | BaseSettings,
) -> set[str]:
    sensitive_fields = [
        field_name
        for field_name, field in model.__fields__.items()
        if field.field_info.extra.get("sensitive", False)
    ]
    sensitive_set = {str(getattr(model, field_name)) for field_name in sensitive_fields}

    recursive_sensitive_data = [
        getattr(model, field_name).get_sensitive_fields_data()
        for field_name, field in model.__fields__.items()
        if isinstance(getattr(model, field_name), BaseOceanModel)
    ]
    for sensitive_data in recursive_sensitive_data:
        sensitive_set.update(sensitive_data)

    return sensitive_set



from abc import abstractmethod

from loguru import logger

from port_ocean.core.handlers.base import BaseHandler
from port_ocean.core.handlers.port_app_config.models import ResourceConfig
from port_ocean.core.ocean_types import (
    RAW_ITEM,
    CalculationResult,
    EntitySelectorDiff,
)


class BaseEntityProcessor(BaseHandler):
    """Abstract base class for processing and parsing entities.

    This class defines abstract methods for parsing raw entity data into entity diffs.

    Attributes:
        context (Any): The context to be used during entity processing.
    """

    @abstractmethod
    async def _parse_items(
        self,
        mapping: ResourceConfig,
        raw_data: list[RAW_ITEM],
        parse_all: bool = False,
        send_raw_data_examples_amount: int = 0,
    ) -> CalculationResult:
        pass

    async def parse_items(
        self,
        mapping: ResourceConfig,
        raw_data: list[RAW_ITEM],
        parse_all: bool = False,
        send_raw_data_examples_amount: int = 0,
    ) -> CalculationResult:
        """Public method to parse raw entity data and map it to an EntityDiff.

        Args:
            mapping (ResourceConfig): The configuration for entity mapping.
            raw_data (list[RawEntity]): The raw data to be parsed.
            parse_all (bool): Whether to parse all data or just data that passed the selector.
            send_raw_data_examples_amount (bool): Whether to send example data to the integration service.

        Returns:
            EntityDiff: The parsed entity differences.
        """
        with logger.contextualize(kind=mapping.kind):
            if not raw_data:
                return CalculationResult(EntitySelectorDiff([], []), [])

            return await self._parse_items(
                mapping, raw_data, parse_all, send_raw_data_examples_amount
            )



from port_ocean.log.handlers import _serialize_record
from loguru import logger
from logging import LogRecord
from queue import Queue
from logging.handlers import QueueHandler
from typing import Callable, Any


log_message = "This is a test log message."
exception_grouop_message = "Test Exception group"
exception_message = "Test Exception"
expected_keys = ["message", "level", "timestamp", "extra"]


def test_serialize_record_log_shape() -> None:
    record = log_record(
        lambda: logger.exception(
            log_message,
            exc_info=None,
        )
    )
    serialized_record = _serialize_record(record)
    assert all(key in serialized_record for key in expected_keys)
    assert log_message in serialized_record.get("message", None)


def test_serialize_record_exc_info_single_exception() -> None:
    record = log_record(
        lambda: logger.exception(
            log_message,
            exc_info=ExceptionGroup(
                exception_grouop_message, [Exception(exception_message)]
            ),
        )
    )
    serialized_record = _serialize_record(record)
    exc_info = assert_extra(serialized_record.get("extra", {}))
    assert exception_grouop_message in exc_info
    assert exception_message in exc_info


def test_serialize_record_exc_info_group_exception() -> None:
    record = log_record(
        lambda: logger.exception(log_message, exc_info=Exception(exception_message))
    )
    serialized_record = _serialize_record(record)
    exc_info = assert_extra(serialized_record.get("extra", {}))
    assert exception_message in exc_info


def assert_extra(extra: dict[str, Any]) -> str:
    exc_info = extra.get("exc_info", None)
    assert type(exc_info) is str
    return exc_info


def log_record(cb: Callable[[], None]) -> LogRecord:
    queue = Queue[LogRecord]()
    queue_handler = QueueHandler(queue)
    logger_id = logger.add(
        queue_handler,
        level="DEBUG",
        format="{message}",
        diagnose=False,
        enqueue=True,
    )
    cb()
    logger.complete()
    logger.remove(logger_id)
    record = queue.get()
    return record



import os
import shutil


def handle_private_integration_flags():
    infra_make_file = "../_infra/Makefile"
    target_link_make_file = os.path.join("./Makefile")

    if "{{ cookiecutter.is_private_integration }}" == "True":
        shutil.copyfile(infra_make_file, target_link_make_file)
        os.remove("sonar-project.properties")
        return

    os.symlink(infra_make_file, target_link_make_file)
    os.remove("Dockerfile")
    os.remove(".dockerignore")


if __name__ == "__main__":
    handle_private_integration_flags()



from typing import Any
from unittest.mock import MagicMock

import pytest

from port_ocean.clients.port.mixins.entities import EntityClientMixin
from port_ocean.core.models import Entity
from httpx import ReadTimeout


errored_entity_identifier: str = "a"
expected_result_entities = [
    Entity(identifier="b", blueprint="b"),
    Entity(identifier="c", blueprint="c"),
]
all_entities = [
    Entity(identifier=errored_entity_identifier, blueprint="a")
] + expected_result_entities


async def mock_upsert_entity(entity: Entity, *args: Any, **kwargs: Any) -> Entity:
    if entity.identifier == errored_entity_identifier:
        raise ReadTimeout("")
    else:
        return entity


@pytest.fixture
async def entity_client(monkeypatch: Any) -> EntityClientMixin:
    # Arrange
    entity_client = EntityClientMixin(auth=MagicMock(), client=MagicMock())
    monkeypatch.setattr(entity_client, "upsert_entity", mock_upsert_entity)

    return entity_client


async def test_batch_upsert_entities_read_timeout_should_raise_false(
    entity_client: EntityClientMixin,
) -> None:
    result_entities = await entity_client.batch_upsert_entities(
        entities=all_entities, request_options=MagicMock(), should_raise=False
    )

    assert result_entities == expected_result_entities


async def test_batch_upsert_entities_read_timeout_should_raise_true(
    entity_client: EntityClientMixin,
) -> None:
    with pytest.raises(ReadTimeout):
        await entity_client.batch_upsert_entities(
            entities=all_entities, request_options=MagicMock(), should_raise=True
        )



from collections import defaultdict

from loguru import logger

from port_ocean.core.ocean_types import (
    IntegrationEventsCallbacks,
    START_EVENT_LISTENER,
    RESYNC_EVENT_LISTENER,
)


class EventsMixin:
    """A mixin class that provides event handling capabilities for the integration class.

    This mixin allows classes to register event listeners and manage event callbacks.
    It provides methods for attaching listeners to "start" and "resync" events.

    Attributes:
        event_strategy (IntegrationEventsCallbacks): A dictionary storing event callbacks.
            - "start": List of functions to be called on "start" event.
            - "resync": Default dictionary mapping event kinds to lists of functions
                        to be called on "resync" events of the specified kind.
    """

    def __init__(self) -> None:
        self.event_strategy: IntegrationEventsCallbacks = {
            "start": [],
            "resync": defaultdict(list),
        }

    @property
    def available_resync_kinds(self) -> list[str]:
        return list(self.event_strategy["resync"].keys())

    def on_start(self, func: START_EVENT_LISTENER) -> START_EVENT_LISTENER:
        """Register a function as a listener for the "start" event.

        Args:
            func (START_EVENT_LISTENER): The function to be called on the "start" event.

        Returns:
            START_EVENT_LISTENER: The input function, unchanged.
        """
        logger.debug(f"Registering {func} as a start event listener")
        self.event_strategy["start"].append(func)
        return func

    def on_resync(
        self, func: RESYNC_EVENT_LISTENER, kind: str | None = None
    ) -> RESYNC_EVENT_LISTENER:
        """Register a function as a listener for a "resync" event.

        Args:
            func (RESYNC_EVENT_LISTENER): The function to be called on the "resync" event.
            kind (str | None, optional): The kind of "resync" event. Defaults to None.

        Returns:
            RESYNC_EVENT_LISTENER: The input function, unchanged.
        """
        if kind is None:
            logger.debug(f"Registering resync event listener any kind")
        else:
            logger.info(f"Registering resync event listener for kind {kind}")
        self.event_strategy["resync"][kind].append(func)
        return func



from abc import abstractmethod
from typing import Type, Any

from loguru import logger
from pydantic import ValidationError

from port_ocean.context.event import event
from port_ocean.context.ocean import PortOceanContext
from port_ocean.core.handlers.base import BaseHandler
from port_ocean.core.handlers.port_app_config.models import PortAppConfig
from port_ocean.utils.misc import get_time


class PortAppConfigCache:
    _port_app_config: PortAppConfig | None
    _retrieval_time: float

    def __init__(self, cache_ttl: int):
        self._port_app_config = None
        self._cache_ttl = cache_ttl

    @property
    def port_app_config(self) -> PortAppConfig:
        if self._port_app_config is None:
            raise ValueError("Port app config is not set")
        return self._port_app_config

    @port_app_config.setter
    def port_app_config(self, value: PortAppConfig) -> None:
        self._retrieval_time = get_time()
        self._port_app_config = value

    @property
    def is_cache_invalid(self) -> bool:
        return (
            not self._port_app_config
            or self._retrieval_time + self._cache_ttl < get_time()
        )


class BasePortAppConfig(BaseHandler):
    """Abstract base class for managing port application configurations.

    This class defines methods for obtaining and processing port application configurations.

    Attributes:
        context (Any): The context to be used during port application configuration.
        CONFIG_CLASS (Type[PortAppConfig]): The class used for defining port application configuration settings.
    """

    CONFIG_CLASS: Type[PortAppConfig] = PortAppConfig

    def __init__(self, context: PortOceanContext):
        super().__init__(context)
        self._app_config_cache = PortAppConfigCache(
            self.context.config.port.port_app_config_cache_ttl
        )

    @abstractmethod
    async def _get_port_app_config(self) -> dict[str, Any]:
        pass

    async def get_port_app_config(self, use_cache: bool = True) -> PortAppConfig:
        """
        Retrieve and parse the port application configuration.

        :param use_cache: Determines whether to use the cached port-app-config if it exists, or to fetch it regardless
        :return: The parsed port application configuration.
        """
        if not use_cache or self._app_config_cache.is_cache_invalid:
            raw_config = await self._get_port_app_config()
            try:
                self._app_config_cache.port_app_config = self.CONFIG_CLASS.parse_obj(
                    raw_config
                )
            except ValidationError:
                logger.error(
                    "Invalid port app config found. Please check that the integration has been configured correctly."
                )
                logger.warning(f"Invalid port app config: {raw_config}")
                raise

        event.port_app_config = self._app_config_cache.port_app_config
        return self._app_config_cache.port_app_config



from abc import abstractmethod
from typing import TypedDict, Callable, Any, Awaitable

from pydantic import Extra

from port_ocean.config.base import BaseOceanModel
from port_ocean.utils.signal import signal_handler
from port_ocean.context.ocean import ocean
from port_ocean.utils.misc import IntegrationStateStatus


class EventListenerEvents(TypedDict):
    """
    A dictionary containing event types and their corresponding event handlers.
    """

    on_resync: Callable[[dict[Any, Any]], Awaitable[None]]


class BaseEventListener:
    def __init__(
        self,
        events: EventListenerEvents,
    ):
        self.events = events

    async def start(self) -> None:
        signal_handler.register(self._stop)
        await self._start()

    @abstractmethod
    async def _start(self) -> None:
        pass

    def _stop(self) -> None:
        """
        Can be used for event listeners that need cleanup before exiting.
        """
        pass

    async def _before_resync(self) -> None:
        """
        Can be used for event listeners that need to perform some action before resync.
        """
        await ocean.app.resync_state_updater.update_before_resync()

    async def _after_resync(self) -> None:
        """
        Can be used for event listeners that need to perform some action after resync.
        """
        await ocean.app.resync_state_updater.update_after_resync()

    async def _on_resync_failure(self, e: Exception) -> None:
        """
        Can be used for event listeners that need to handle resync failures.
        """
        await ocean.app.resync_state_updater.update_after_resync(
            IntegrationStateStatus.Failed
        )

    async def _resync(
        self,
        resync_args: dict[Any, Any],
    ) -> None:
        """
        Triggers the "on_resync" event.
        """
        await self._before_resync()
        try:
            await self.events["on_resync"](resync_args)
            await self._after_resync()
        except Exception as e:
            await self._on_resync_failure(e)
            raise e


class EventListenerSettings(BaseOceanModel, extra=Extra.allow):
    type: str

    def to_request(self) -> dict[str, Any]:
        """
        Converts the Settings object to a dictionary representation (request format).
        """
        return {"type": self.type}



import asyncio
import random
import time
from datetime import datetime
from functools import partial
from http import HTTPStatus
from typing import Any, Callable, Coroutine, Iterable, Mapping, Union

import httpx
from dateutil.parser import isoparse


# Adapted from https://github.com/encode/httpx/issues/108#issuecomment-1434439481
class RetryTransport(httpx.AsyncBaseTransport, httpx.BaseTransport):
    """
    A custom HTTP transport that automatically retries requests using an exponential backoff strategy
    for specific HTTP status codes and request methods.

    Args:
        wrapped_transport (Union[httpx.BaseTransport, httpx.AsyncBaseTransport]): The underlying HTTP transport
            to wrap and use for making requests.
        max_attempts (int, optional): The maximum number of times to retry a request before giving up. Defaults to 10.
        max_backoff_wait (float, optional): The maximum time to wait between retries in seconds. Defaults to 60.
        backoff_factor (float, optional): The factor by which the wait time increases with each retry attempt.
            Defaults to 0.1.
        jitter_ratio (float, optional): The amount of jitter to add to the backoff time. Jitter is a random
            value added to the backoff time to avoid a "thundering herd" effect. The value should be between 0 and 0.5.
            Defaults to 0.1.
        respect_retry_after_header (bool, optional): Whether to respect the Retry-After header in HTTP responses
            when deciding how long to wait before retrying. Defaults to True.
        retryable_methods (Iterable[str], optional): The HTTP methods that can be retried. Defaults to
            ["HEAD", "GET", "PUT", "DELETE", "OPTIONS", "TRACE"].
        retry_status_codes (Iterable[int], optional): The HTTP status codes that can be retried. Defaults to
            [429, 502, 503, 504].

    Attributes:
        _wrapped_transport (Union[httpx.BaseTransport, httpx.AsyncBaseTransport]): The underlying HTTP transport
            being wrapped.
        _max_attempts (int): The maximum number of times to retry a request.
        _backoff_factor (float): The factor by which the wait time increases with each retry attempt.
        _respect_retry_after_header (bool): Whether to respect the Retry-After header in HTTP responses.
        _retryable_methods (frozenset): The HTTP methods that can be retried.
        _retry_status_codes (frozenset): The HTTP status codes that can be retried.
        _jitter_ratio (float): The amount of jitter to add to the backoff time.
        _max_backoff_wait (float): The maximum time to wait between retries in seconds.

    """

    RETRYABLE_METHODS = frozenset(["HEAD", "GET", "PUT", "DELETE", "OPTIONS", "TRACE"])
    RETRYABLE_STATUS_CODES = frozenset(
        [
            HTTPStatus.TOO_MANY_REQUESTS,
            HTTPStatus.BAD_GATEWAY,
            HTTPStatus.SERVICE_UNAVAILABLE,
            HTTPStatus.GATEWAY_TIMEOUT,
        ]
    )
    MAX_BACKOFF_WAIT = 60

    def __init__(
        self,
        wrapped_transport: Union[httpx.BaseTransport, httpx.AsyncBaseTransport],
        max_attempts: int = 10,
        max_backoff_wait: float = MAX_BACKOFF_WAIT,
        backoff_factor: float = 0.1,
        jitter_ratio: float = 0.1,
        respect_retry_after_header: bool = True,
        retryable_methods: Iterable[str] | None = None,
        retry_status_codes: Iterable[int] | None = None,
        logger: Any | None = None,
    ) -> None:
        """
        Initializes the instance of RetryTransport class with the given parameters.

        Args:
            wrapped_transport (Union[httpx.BaseTransport, httpx.AsyncBaseTransport]):
                The transport layer that will be wrapped and retried upon failure.
            max_attempts (int, optional):
                The maximum number of times the request can be retried in case of failure.
                Defaults to 10.
            max_backoff_wait (float, optional):
                The maximum amount of time (in seconds) to wait before retrying a request.
                Defaults to 60.
            backoff_factor (float, optional):
                The factor by which the waiting time will be multiplied in each retry attempt.
                Defaults to 0.1.
            jitter_ratio (float, optional):
                The ratio of randomness added to the waiting time to prevent simultaneous retries.
                Should be between 0 and 0.5. Defaults to 0.1.
            respect_retry_after_header (bool, optional):
                A flag to indicate if the Retry-After header should be respected.
                If True, the waiting time specified in Retry-After header is used for the waiting time.
                Defaults to True.
            retryable_methods (Iterable[str], optional):
                The HTTP methods that can be retried. Defaults to ['HEAD', 'GET', 'PUT', 'DELETE', 'OPTIONS', 'TRACE'].
            retry_status_codes (Iterable[int], optional):
                The HTTP status codes that can be retried.
                Defaults to [429, 502, 503, 504].
            logger (Any): The logger to use for logging retries.
        """
        self._wrapped_transport = wrapped_transport
        if jitter_ratio < 0 or jitter_ratio > 0.5:
            raise ValueError(
                f"Jitter ratio should be between 0 and 0.5, actual {jitter_ratio}"
            )

        self._max_attempts = max_attempts
        self._backoff_factor = backoff_factor
        self._respect_retry_after_header = respect_retry_after_header
        self._retryable_methods = (
            frozenset(retryable_methods)
            if retryable_methods
            else self.RETRYABLE_METHODS
        )
        self._retry_status_codes = (
            frozenset(retry_status_codes)
            if retry_status_codes
            else self.RETRYABLE_STATUS_CODES
        )
        self._jitter_ratio = jitter_ratio
        self._max_backoff_wait = max_backoff_wait
        self._logger = logger

    def handle_request(self, request: httpx.Request) -> httpx.Response:
        """
        Sends an HTTP request, possibly with retries.

        Args:
            request (httpx.Request): The request to send.

        Returns:
            httpx.Response: The response received.

        """
        transport: httpx.BaseTransport = self._wrapped_transport  # type: ignore
        if request.method in self._retryable_methods:
            send_method = partial(transport.handle_request)
            response = self._retry_operation(request, send_method)
        else:
            response = transport.handle_request(request)
        return response

    async def handle_async_request(self, request: httpx.Request) -> httpx.Response:
        """Sends an HTTP request, possibly with retries.

        Args:
            request: The request to perform.

        Returns:
            The response.

        """
        transport: httpx.AsyncBaseTransport = self._wrapped_transport  # type: ignore
        if self._is_retryable_method(request):
            send_method = partial(transport.handle_async_request)
            response = await self._retry_operation_async(request, send_method)
        else:
            response = await transport.handle_async_request(request)
        return response

    async def aclose(self) -> None:
        """
        Closes the underlying HTTP transport, terminating all outstanding connections and rejecting any further
        requests.

        This should be called before the object is dereferenced, to ensure that connections are properly cleaned up.
        """
        transport: httpx.AsyncBaseTransport = self._wrapped_transport  # type: ignore
        await transport.aclose()

    def close(self) -> None:
        """
        Closes the underlying HTTP transport, terminating all outstanding connections and rejecting any further
        requests.

        This should be called before the object is dereferenced, to ensure that connections are properly cleaned up.
        """
        transport: httpx.BaseTransport = self._wrapped_transport  # type: ignore
        transport.close()

    def _is_retryable_method(self, request: httpx.Request) -> bool:
        return request.method in self._retryable_methods or request.extensions.get(
            "retryable", False
        )

    def _should_retry(self, response: httpx.Response) -> bool:
        return response.status_code in self._retry_status_codes

    def _log_error(
        self,
        request: httpx.Request,
        error: Exception | None,
    ) -> None:
        if not self._logger:
            return

        if isinstance(error, httpx.ConnectTimeout):
            self._logger.error(
                f"Request {request.method} {request.url} failed to connect: {str(error)}"
            )
        elif isinstance(error, httpx.TimeoutException):
            self._logger.error(
                f"Request {request.method} {request.url} failed with a timeout exception: {str(error)}"
            )
        elif isinstance(error, httpx.HTTPError):
            self._logger.error(
                f"Request {request.method} {request.url} failed with an HTTP error: {str(error)}"
            )

    def _log_before_retry(
        self,
        request: httpx.Request,
        sleep_time: float,
        response: httpx.Response | None,
        error: Exception | None,
    ) -> None:
        if self._logger and response:
            self._logger.warning(
                f"Request {request.method} {request.url} failed with status code:"
                f" {response.status_code}, retrying in {sleep_time} seconds."  # noqa: F821
            )
        elif self._logger and error:
            self._logger.warning(
                f"Request {request.method} {request.url} failed with exception:"
                f" {type(error).__name__} - {str(error) or 'No error message'}, retrying in {sleep_time} seconds."
            )

    async def _should_retry_async(self, response: httpx.Response) -> bool:
        return response.status_code in self._retry_status_codes

    def _calculate_sleep(
        self, attempts_made: int, headers: Union[httpx.Headers, Mapping[str, str]]
    ) -> float:
        # Retry-After
        # The Retry-After response HTTP header indicates how long the user agent should wait before
        # making a follow-up request. There are three main cases this header is used:
        # - When sent with a 503 (Service Unavailable) response, this indicates how long the service
        #   is expected to be unavailable.
        # - When sent with a 429 (Too Many Requests) response, this indicates how long to wait before
        #   making a new request.
        # - When sent with a redirect response, such as 301 (Moved Permanently), this indicates the
        #   minimum time that the user agent is asked to wait before issuing the redirected request.
        retry_after_header = (headers.get("Retry-After") or "").strip()
        if self._respect_retry_after_header and retry_after_header:
            if retry_after_header.isdigit():
                return float(retry_after_header)

            try:
                parsed_date = isoparse(
                    retry_after_header
                ).astimezone()  # converts to local time
                diff = (parsed_date - datetime.now().astimezone()).total_seconds()
                if diff > 0:
                    return min(diff, self._max_backoff_wait)
            except ValueError:
                pass

        backoff = self._backoff_factor * (2 ** (attempts_made - 1))
        jitter = (backoff * self._jitter_ratio) * random.choice([1, -1])
        total_backoff = backoff + jitter
        return min(total_backoff, self._max_backoff_wait)

    async def _retry_operation_async(
        self,
        request: httpx.Request,
        send_method: Callable[..., Coroutine[Any, Any, httpx.Response]],
    ) -> httpx.Response:
        remaining_attempts = self._max_attempts
        attempts_made = 0
        response: httpx.Response | None = None
        error: Exception | None = None
        while True:
            if attempts_made > 0:
                sleep_time = self._calculate_sleep(attempts_made, {})
                self._log_before_retry(request, sleep_time, response, error)
                await asyncio.sleep(sleep_time)

            error = None
            response = None
            try:
                response = await send_method(request)
                response.request = request
                if remaining_attempts < 1 or not (
                    await self._should_retry_async(response)
                ):
                    return response
                await response.aclose()
            except httpx.ConnectTimeout as e:
                error = e
                if remaining_attempts < 1:
                    self._log_error(request, error)
                    raise
            except httpx.ReadTimeout as e:
                error = e
                if remaining_attempts < 1:
                    self._log_error(request, error)
                    raise
            except httpx.TimeoutException as e:
                error = e
                if remaining_attempts < 1:
                    self._log_error(request, error)
                    raise
            except httpx.HTTPError as e:
                error = e
                if remaining_attempts < 1:
                    self._log_error(request, error)
                    raise
            attempts_made += 1
            remaining_attempts -= 1

    def _retry_operation(
        self,
        request: httpx.Request,
        send_method: Callable[..., httpx.Response],
    ) -> httpx.Response:
        remaining_attempts = self._max_attempts
        attempts_made = 0
        response: httpx.Response | None = None
        error: Exception | None = None
        while True:
            if attempts_made > 0:
                sleep_time = self._calculate_sleep(attempts_made, {})
                self._log_before_retry(request, sleep_time, response, error)
                time.sleep(sleep_time)

            error = None
            response = None
            try:
                response = send_method(request)
                response.request = request
                if remaining_attempts < 1 or not self._should_retry(response):
                    return response
                response.close()
            except httpx.ConnectTimeout as e:
                error = e
                if remaining_attempts < 1:
                    self._log_error(request, error)
                    raise
            except httpx.TimeoutException as e:
                error = e
                if remaining_attempts < 1:
                    self._log_error(request, error)
                    raise
            except httpx.HTTPError as e:
                error = e
                if remaining_attempts < 1:
                    self._log_error(request, error)
                    raise
            attempts_made += 1
            remaining_attempts -= 1



from enum import Enum
import inspect
from importlib.util import spec_from_file_location, module_from_spec
from pathlib import Path
from time import time
from types import ModuleType
from typing import Callable, Any
from uuid import uuid4
import tomli
import yaml


class IntegrationStateStatus(Enum):
    Running = "running"
    Failed = "failed"
    Completed = "completed"


def get_time(seconds_precision: bool = True) -> float:
    """Return current time as Unix/Epoch timestamp, in seconds.
    :param seconds_precision: if True, return with seconds precision as integer (default).
                              If False, return with milliseconds precision as floating point number of seconds.
    """
    return time() if not seconds_precision else int(time())


def generate_uuid() -> str:
    """Return a UUID4 as string"""
    return str(uuid4())


def get_function_location(func: Callable[..., Any]) -> str:
    file_path = inspect.getsourcefile(func)
    line_number = inspect.getsourcelines(func)[1]
    return f"{file_path}:{line_number}"


def get_pyproject_data() -> dict[str, Any] | None:
    try:
        with open("./pyproject.toml", "rb") as toml_file:
            pyproject_data = tomli.load(toml_file)
            return pyproject_data["tool"]["poetry"]
    except (FileNotFoundError, KeyError):
        return None


def get_integration_version() -> str:
    if data := get_pyproject_data():
        return data["version"]
    return ""


def get_integration_name() -> str:
    if data := get_pyproject_data():
        return data["name"]
    return ""


def get_spec_file(path: Path = Path(".")) -> dict[str, Any] | None:
    try:
        return yaml.safe_load((path / ".port/spec.yaml").read_text())
    except FileNotFoundError:
        return None


def load_module(file_path: str) -> ModuleType:
    spec = spec_from_file_location("module.name", file_path)
    if spec is None or spec.loader is None:
        raise Exception(f"Failed to load integration from path: {file_path}")

    module = module_from_spec(spec)
    spec.loader.exec_module(module)

    return module



from __future__ import annotations

from typing import Any

from pydantic import BaseModel, Field

from port_ocean.clients.port.types import RequestOptions


class Rule(BaseModel):
    property: str
    operator: str
    value: str


class IngestSearchQuery(BaseModel):
    combinator: str
    rules: list[Rule | IngestSearchQuery]


class EntityMapping(BaseModel):
    identifier: str | IngestSearchQuery
    title: str | None
    blueprint: str
    team: str | None
    properties: dict[str, str] = Field(default_factory=dict)
    relations: dict[str, str | IngestSearchQuery] = Field(default_factory=dict)

    @property
    def is_using_search_identifier(self) -> bool:
        return isinstance(self.identifier, dict)


class MappingsConfig(BaseModel):
    mappings: EntityMapping


class PortResourceConfig(BaseModel):
    entity: MappingsConfig
    items_to_parse: str | None = Field(alias="itemsToParse")


class Selector(BaseModel):
    query: str


class ResourceConfig(BaseModel):
    kind: str
    selector: Selector
    port: PortResourceConfig


class PortAppConfig(BaseModel):
    enable_merge_entity: bool = Field(alias="enableMergeEntity", default=True)
    delete_dependent_entities: bool = Field(
        alias="deleteDependentEntities", default=True
    )
    create_missing_related_entities: bool = Field(
        alias="createMissingRelatedEntities", default=True
    )
    resources: list[ResourceConfig] = Field(default_factory=list)

    def get_port_request_options(self) -> RequestOptions:
        return {
            "delete_dependent_entities": self.delete_dependent_entities,
            "create_missing_related_entities": self.create_missing_related_entities,
            "merge": self.enable_merge_entity,
            "validation_only": False,
        }

    def to_request(self) -> dict[str, Any]:
        return {
            "deleteDependentEntities": self.delete_dependent_entities,
            "createMissingRelatedEntities": self.create_missing_related_entities,
            "enableMergeEntity": self.enable_merge_entity,
            "resources": [
                resource.dict(by_alias=True, exclude_none=True, exclude_unset=True)
                for resource in self.resources
            ],
        }

    class Config:
        allow_population_by_field_name = True
        validate_assignment = True



from loguru import logger

from port_ocean.clients.port.authentication import PortAuthentication
from port_ocean.clients.port.mixins.blueprints import BlueprintClientMixin
from port_ocean.clients.port.mixins.entities import EntityClientMixin
from port_ocean.clients.port.mixins.integrations import IntegrationClientMixin
from port_ocean.clients.port.mixins.migrations import MigrationClientMixin
from port_ocean.clients.port.types import (
    KafkaCreds,
)
from port_ocean.clients.port.utils import (
    handle_status_code,
    get_internal_http_client,
)
from port_ocean.exceptions.clients import KafkaCredentialsNotFound
from typing import Any


class PortClient(
    EntityClientMixin,
    IntegrationClientMixin,
    BlueprintClientMixin,
    MigrationClientMixin,
):
    def __init__(
        self,
        base_url: str,
        client_id: str,
        client_secret: str,
        integration_identifier: str,
        integration_type: str,
        integration_version: str,
    ):
        self.api_url = f"{base_url}/v1"
        self.client = get_internal_http_client(self)
        self.auth = PortAuthentication(
            self.client,
            client_id,
            client_secret,
            self.api_url,
            integration_identifier,
            integration_type,
            integration_version,
        )
        EntityClientMixin.__init__(self, self.auth, self.client)
        IntegrationClientMixin.__init__(
            self, integration_identifier, integration_version, self.auth, self.client
        )
        BlueprintClientMixin.__init__(self, self.auth, self.client)
        MigrationClientMixin.__init__(self, self.auth, self.client)

    async def get_kafka_creds(self) -> KafkaCreds:
        logger.info("Fetching organization kafka credentials")
        response = await self.client.get(
            f"{self.api_url}/kafka-credentials", headers=await self.auth.headers()
        )
        if response.is_error:
            logger.error("Error getting kafka credentials")
        handle_status_code(response)

        credentials = response.json().get("credentials")

        if credentials is None:
            raise KafkaCredentialsNotFound("No kafka credentials found")

        return credentials

    async def get_org_id(self) -> str:
        logger.info("Fetching organization id")

        response = await self.client.get(
            f"{self.api_url}/organization", headers=await self.auth.headers()
        )
        if response.is_error:
            logger.error(f"Error getting organization id, error: {response.text}")
        handle_status_code(response)

        return response.json()["organization"]["id"]

    async def update_integration_state(
        self, state: dict[str, Any], should_raise: bool = True, should_log: bool = True
    ) -> dict[str, Any]:
        if should_log:
            logger.debug(f"Updating integration resync state with: {state}")
        response = await self.client.patch(
            f"{self.api_url}/integration/{self.integration_identifier}/resync-state",
            headers=await self.auth.headers(),
            json=state,
        )
        handle_status_code(response, should_raise, should_log)
        if response.is_success and should_log:
            logger.info("Integration resync state updated successfully")

        return response.json().get("integration", {})



import json
from pathlib import Path
from typing import Type, Any, TypedDict, Optional

import httpx
import yaml
from pydantic import BaseModel, Field
from starlette import status

from port_ocean.clients.port.client import PortClient
from port_ocean.core.handlers.port_app_config.models import PortAppConfig
from port_ocean.exceptions.port_defaults import (
    UnsupportedDefaultFileType,
)

YAML_EXTENSIONS = [".yaml", ".yml"]
ALLOWED_FILE_TYPES = [".json", *YAML_EXTENSIONS]


class Preset(TypedDict):
    blueprint: str
    data: list[dict[str, Any]]


class Defaults(BaseModel):
    blueprints: list[dict[str, Any]] = []
    actions: list[dict[str, Any]] = []
    scorecards: list[Preset] = []
    pages: list[dict[str, Any]] = []
    port_app_config: Optional[PortAppConfig] = Field(
        default=None, alias="port-app-config"
    )

    class Config:
        allow_population_by_field_name = True


async def is_integration_exists(port_client: PortClient) -> bool:
    try:
        await port_client.get_current_integration(should_log=False)
        return True
    except httpx.HTTPStatusError as e:
        if e.response.status_code != status.HTTP_404_NOT_FOUND:
            raise e

    return False


def deconstruct_blueprints_to_creation_steps(
    raw_blueprints: list[dict[str, Any]]
) -> tuple[list[dict[str, Any]], ...]:
    """
    Deconstructing the blueprint into stages so the api wont fail to create a blueprint if there is a conflict
    example: Preventing the failure of creating a blueprint with a relation to another blueprint
    """
    (
        bare_blueprint,
        with_relations,
        full_blueprint,
    ) = ([], [], [])

    for blueprint in raw_blueprints.copy():
        full_blueprint.append(blueprint.copy())

        blueprint.pop("calculationProperties", {})
        blueprint.pop("mirrorProperties", {})
        with_relations.append(blueprint.copy())

        blueprint.pop("teamInheritance", {})
        blueprint.pop("relations", {})
        bare_blueprint.append(blueprint)

    return (
        bare_blueprint,
        with_relations,
        full_blueprint,
    )


def get_port_integration_defaults(
    port_app_config_class: Type[PortAppConfig], base_path: Path = Path(".")
) -> Defaults | None:
    defaults_dir = base_path / ".port/resources"
    if not defaults_dir.exists():
        return None

    if not defaults_dir.is_dir():
        raise UnsupportedDefaultFileType(
            f"Defaults directory is not a directory: {defaults_dir}"
        )

    default_jsons = {}
    allowed_file_names = [
        field_model.alias for _, field_model in Defaults.__fields__.items()
    ]
    for path in defaults_dir.iterdir():
        if path.stem in allowed_file_names:
            if not path.is_file() or path.suffix not in ALLOWED_FILE_TYPES:
                raise UnsupportedDefaultFileType(
                    f"Defaults directory should contain only one of the next types: {ALLOWED_FILE_TYPES}. Found: {path}"
                )

            if path.suffix in YAML_EXTENSIONS:
                default_jsons[path.stem] = yaml.safe_load(path.read_text())
            else:
                default_jsons[path.stem] = json.loads(path.read_text())

    return Defaults(
        blueprints=default_jsons.get("blueprints", []),
        actions=default_jsons.get("actions", []),
        scorecards=default_jsons.get("scorecards", []),
        pages=default_jsons.get("pages", []),
        port_app_config=port_app_config_class(
            **default_jsons.get("port-app-config", {})
        ),
    )
